# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from unittest.mock import patch

import networkx as nx
import pytest

import lisien.examples.kobold as kobold
import lisien.examples.polygons as polygons
from lisien.engine import Engine
from lisien.proxy import EngineProcessManager
from lisien.proxy.handle import EngineHandle
from lisien.tests import data


def test_fast_delta(handle_initialized):
	hand = handle_initialized
	unpack_delta = hand._real._unpack_slightly_packed_delta

	# there's currently no way to do fast delta past the time when
	# a character was created, due to the way keyframes work...
	# so don't test that

	branch, turn, tick = hand._real._btt()
	ret, diff = hand.next_turn()
	btt = hand._real._btt()
	slowd = unpack_delta(
		hand._get_slow_delta(btt_from=(branch, turn, tick), btt_to=btt)
	)
	fastd = hand.unpack(diff)
	if "universal" in slowd:
		del slowd["universal"]["rando_state"]
		del fastd["universal"]["rando_state"]
	assert fastd == slowd, "Fast delta differs from slow delta"
	ret, diff2 = hand.time_travel("trunk", 0, tick)
	btt2 = hand._real._btt()
	slowd2 = unpack_delta(hand._get_slow_delta(btt_from=btt, btt_to=btt2))
	fastd2 = hand.unpack(diff2)
	if "universal" in slowd:
		del slowd2["universal"]["rando_state"]
		del fastd2["universal"]["rando_state"]
	assert fastd2 == slowd2, "Fast delta differs from slow delta"
	ret, diff3 = hand.time_travel("trunk", 1)
	btt3 = hand._real._btt()
	slowd3 = unpack_delta(hand._get_slow_delta(btt_from=btt2, btt_to=btt3))
	fastd3 = hand.unpack(diff3)
	if "universal" in slowd:
		del slowd3["universal"]["rando_state"]
		del fastd3["universal"]["rando_state"]
	assert fastd3 == slowd3, "Fast delta differs from slow delta"


def test_serialize_deleted(college24_premade):
	eng = college24_premade
	d0r0s0 = eng.character["dorm0room0student0"]
	roommate = d0r0s0.stat["roommate"]
	del eng.character[roommate.name]
	assert not roommate
	with pytest.raises(KeyError):
		eng.character[roommate.name]
	assert d0r0s0.stat["roommate"] == roommate
	assert eng.unpack(eng.pack(d0r0s0.stat["roommate"])) == roommate


def test_manip_deleted(sqleng):
	eng = sqleng
	phys = eng.new_character("physical")
	phys.stat["aoeu"] = True
	phys.add_node(0)
	phys.add_node(1)
	phys.node[1]["aoeu"] = True
	del phys.node[1]
	phys.add_node(1)
	assert "aoeu" not in phys.node[1]
	phys.add_edge(0, 1)
	phys.adj[0][1]["aoeu"] = True
	del phys.adj[0][1]
	phys.add_edge(0, 1)
	assert "aoeu" not in phys.adj[0][1]
	del eng.character["physical"]
	assert not phys
	phys = eng.new_character("physical")
	assert "aoeu" not in phys.stat
	assert 0 not in phys
	assert 1 not in phys
	assert 0 not in phys.adj
	assert 1 not in phys.adj


def test_switch_main_branch(engine):
	phys = engine.new_character("physical", hello="hi")
	engine.next_turn()
	phys.stat["hi"] = "hello"
	with pytest.raises(ValueError):
		engine.switch_main_branch("tronc")
	engine.turn = 0
	engine.tick = 0
	engine.switch_main_branch("tronc")
	assert engine.branch == "tronc"
	assert phys
	assert "hi" not in phys.stat
	assert "hello" in phys.stat
	engine.next_turn()
	phys.stat["hi"] = "hey there"
	engine.turn = 0
	engine.tick = 0
	engine.switch_main_branch("trunk")
	assert phys.stat["hello"] == "hi"
	engine.turn = 1
	assert phys.stat["hello"] == "hi"
	assert phys.stat["hi"] == "hello"


def test_updnoderb(handle):
	engine = handle._real
	char0 = engine.new_character("0")
	node0 = char0.new_place("0")

	@node0.rule(always=True)
	def change_rulebook(node):
		node.rulebook = "haha"

	a, b = handle.next_turn()

	delta = engine.unpack(b)

	assert delta
	assert "0" in delta
	assert "node_val" in delta["0"]
	assert "0" in delta["0"]["node_val"]
	assert "0" in delta["0"]["node_val"]
	assert "rulebook" in delta["0"]["node_val"]["0"]
	assert delta["0"]["node_val"]["0"]["rulebook"] == "haha"


def test_updedgerb(handle):
	engine = handle._real
	char0 = engine.new_character("0")
	node0 = char0.new_place("0")
	node1 = char0.new_place("1")
	edge = node0.new_portal(node1)

	@edge.rule(always=True)
	def change_rulebook(edge):
		edge.rulebook = "haha"

	a, b = handle.next_turn()

	delta = engine.unpack(b)

	assert (
		"0" in delta
		and "edge_val" in delta["0"]
		and "0" in delta["0"]["edge_val"]
		and "1" in delta["0"]["edge_val"]["0"]
		and "rulebook" in delta["0"]["edge_val"]["0"]["1"]
		and delta["0"]["edge_val"]["0"]["1"]["rulebook"] == "haha"
	)


def test_thing_place_iter(tmp_path):
	# set up some world state with things and places, before starting the proxy
	with Engine(tmp_path, workers=0) as eng:
		kobold.inittest(eng)
	manager = EngineProcessManager()
	engine = manager.start(tmp_path, workers=0)
	phys = engine.character["physical"]
	for place_name in phys.place:
		assert isinstance(place_name, tuple)
	for thing_name in phys.thing:
		assert isinstance(thing_name, str)
	manager.shutdown()


@pytest.fixture
def mocked_keyframe(tmp_path):
	with (
		patch("lisien.Engine.snap_keyframe"),
		Engine(
			tmp_path,
			random_seed=69105,
			enforce_end_of_time=False,
			keyframe_on_close=False,
			workers=0,
		) as eng,
	):
		for _ in range(8):
			eng.next_turn()
		eng._start_branch("trunk", *data.BTT_FROM)
		eng.snap_keyframe.side_effect = [data.KF_FROM, data.KF_TO]
		eng._set_btt(*data.BTT_FROM)
		yield eng


@pytest.mark.parametrize("run", list(range(10)))
def test_get_slow_delta_overload(mocked_keyframe, run):
	eng = mocked_keyframe
	slowd = eng._unpack_slightly_packed_delta(
		eng._get_slow_delta(data.BTT_FROM, data.BTT_TO)
	)
	assert slowd == data.SLOW_DELTA


@pytest.mark.parametrize("algorithm", ["slow", "fast"])
def test_apply_delta(tmp_path, algorithm):
	slow = algorithm == "slow"
	with Engine(tmp_path, workers=0) as eng:
		initial_state = nx.DiGraph(
			{
				0: {1: {"omg": "lol"}},
				1: {0: {"omg": "blasphemy"}},
				2: {},
				3: {},
				"it": {},
			}
		)
		initial_state.nodes()[2]["hi"] = "hello"
		initial_state.nodes()["it"]["location"] = 0
		initial_state.graph["wat"] = "nope"
		phys = eng.new_character("physical", initial_state)
		eng.add_character("pointless")
		kf0 = eng.snap_keyframe()
		del kf0["universal"]["rando_state"]
		if slow:
			eng.branch = "b"
		else:
			eng.next_turn()
		kf1 = eng.snap_keyframe()
		del kf1["universal"]["rando_state"]
		assert kf0 == kf1
		del phys.portal[1][0]
		port = phys.new_portal(0, 2)
		port["hi"] = "bye"
		phys.place[1]["wtf"] = "bbq"
		phys.thing["it"].location = phys.place[1]
		del phys.place[3]
		eng.add_character("pointed")
		del eng.character["pointless"]
		assert "pointless" not in eng.character, "Failed to delete character"
		phys.portal[0][1]["meaning"] = 42
		del phys.portal[0][1]["omg"]
		if slow:
			eng.branch = "trunk"
		else:
			eng.turn = 0
			eng.tick = 0
	mang = EngineProcessManager()
	try:
		prox = mang.start(tmp_path, workers=0)
		assert prox.turn == 0
		phys = prox.character["physical"]
		assert 3 in phys.place
		assert phys.portal[1][0]["omg"] == "blasphemy"
		assert "it" in phys.thing
		if slow:
			prox.branch = "b"
		else:
			prox.turn = 1
		assert 3 not in phys.place
		assert not list(phys.portal[1])
		assert 2 in phys.portal[0]
		assert phys.portal[0][2]["hi"] == "bye"
		assert phys.place[1]["wtf"] == "bbq"
		assert phys.thing["it"].location == phys.place[1]
		assert "pointless" not in prox.character, "Loaded deleted character"
		assert "pointed" in prox.character
		assert phys.portal[0][1]["meaning"] == 42
		assert "omg" not in phys.portal[0][1]
	finally:
		mang.shutdown()


@pytest.fixture
def polys(tmp_path):
	with Engine(tmp_path, workers=0, random_seed=69105) as eng:
		polygons.install(eng)
	return tmp_path


def test_change_triggers(polys):
	procman = EngineProcessManager()
	eng = procman.start(polys)
	relocate = eng.character["triangle"].unit.rule["relocate"]
	assert list(relocate.triggers) == [
		eng.trigger.similar_neighbors,
		eng.trigger.dissimilar_neighbors,
	]
	relocate.triggers = ["dissimilar_neighbors"]
	procman.shutdown()
	with Engine(polys, workers=0) as eng:
		assert list(
			eng.character["triangle"].unit.rule["relocate"].triggers
		) == [eng.trigger.dissimilar_neighbors]


def test_change_string(tmp_path):
	handle = EngineHandle(tmp_path, workers=0)
	handle.set_language("eng")
	handle.set_string("a string", "its value")
	assert handle.get_string_lang_items("eng") == [("a string", "its value")]
	handle.close()
	with Engine(tmp_path, workers=0) as eng:
		assert eng.string["a string"] == "its value"



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Tests for the rules engine's basic polling functionality

Make sure that every type of rule gets followed, and that the fact
it was followed got recorded correctly.

"""


def test_character_dot_rule(engine):
	"""Test that a rule on a character is polled correctly"""
	eng = engine
	char = eng.new_character("who")

	@char.rule(always=True)
	def yes(char):
		char.stat["run"] = True

	eng.next_turn()
	btt = eng._btt()
	assert char.stat["run"]
	eng.time = "trunk", 0, 0
	assert "run" not in char.stat
	eng.next_turn()
	assert btt == eng._btt()
	assert char.stat["run"]


def test_unit_dot_rule(engine):
	"""Test that a rule applied to a character's avatars is polled correctly"""
	char = engine.new_character("char")
	graph = engine.new_character("graph")
	av = graph.new_place("av")
	char.add_unit(av)
	starttick = engine.tick

	@char.unit.rule(always=True)
	def yes(av):
		av["run"] = True

	engine.next_turn()
	btt = engine._btt()
	assert av["run"]
	engine.time = "trunk", 0, starttick
	assert "run" not in av
	engine.next_turn()
	assert btt == engine._btt()
	assert av["run"]


def test_thing_dot_rule(engine):
	"""Test that a rule applied to a thing mapping is polled correctly"""
	char = engine.new_character("char")
	place = char.new_place("place")
	thing = place.new_thing("thing")
	starttick = engine.tick

	@char.thing.rule(always=True)
	def yes(thing):
		thing["run"] = True

	engine.next_turn()
	btt = engine._btt()
	assert thing["run"]
	engine.time = "trunk", 0, starttick
	assert "run" not in thing
	engine.next_turn()
	assert btt == engine._btt()
	assert thing["run"]


def test_place_dot_rule(engine):
	"""Test that a rule applied to a place mapping is polled correctly"""
	char = engine.new_character("char")
	place = char.new_place("place")
	starttick = engine.tick

	@char.place.rule(always=True)
	def yes(plac):
		plac["run"] = True

	engine.next_turn()
	btt = engine._btt()
	assert place["run"]
	engine.time = "trunk", 0, starttick
	assert "run" not in place
	engine.next_turn()
	assert btt == engine._btt()
	assert place["run"]


def test_portal_dot_rule(engine):
	"""Test that a rule applied to a portal mapping is polled correctly"""
	char = engine.new_character("char")
	orig = char.new_place("orig")
	dest = char.new_place("dest")
	port = orig.new_portal(dest)
	starttick = engine.tick

	@char.portal.rule(always=True)
	def yes(portl):
		portl["run"] = True

	engine.next_turn()
	btt = engine._btt()
	assert port["run"]
	engine.time = "trunk", 0, starttick
	assert "run" not in port
	engine.next_turn()
	assert btt == engine._btt()
	assert port["run"]


def test_node_rule(engine):
	"""Test that a rule applied to one node is polled correctly"""
	char = engine.new_character("char")
	place = char.new_place("place")
	thing = place.new_thing("thing")
	starttick = engine.tick

	@place.rule(always=True)
	def yes(plac):
		plac["run"] = True

	@thing.rule(always=True)
	def definitely(thig):
		thig["run"] = True

	engine.next_turn()
	btt = engine._btt()
	assert place["run"]
	assert thing["run"]
	engine.time = "trunk", 0, starttick
	assert "run" not in place
	assert "run" not in thing
	engine.next_turn()
	assert btt == engine._btt()
	assert place["run"]
	assert thing["run"]


def test_portal_rule(engine):
	"""Test that a rule applied to one portal is polled correctly"""
	char = engine.new_character("char")
	orig = char.new_place("orig")
	dest = char.new_place("dest")
	port = orig.new_portal(dest)
	starttick = engine.tick

	@port.rule(always=True)
	def yes(portl):
		portl["run"] = True

	engine.next_turn()
	btt = engine._btt()
	assert port["run"]
	engine.time = "trunk", 0, starttick
	assert "run" not in port
	engine.next_turn()
	assert btt == engine._btt()
	assert port["run"]


def test_post_time_travel_increment(engine):
	"""Test that, when the rules are run after time travel resulting in
	a tick greater than zero, we advance to the next turn before running rules

	"""
	char = engine.new_character("char")
	char.stat["something"] = 0
	place = char.new_place("there")
	place["otherthing"] = 0

	@char.rule(always=True)
	def incr(chara):
		chara.stat["something"] += 1

	@place.rule(always=True)
	def decr(plac):
		plac["otherthing"] -= 1

	engine.next_turn()
	engine.next_turn()
	assert engine.tick == 2
	engine.branch = "branch1"
	assert engine.tick == 2
	engine.next_turn()
	assert engine.tick == 2
	engine.turn = 2
	engine.branch = "trunk"
	assert engine.tick == 2
	engine.next_turn()
	assert engine.tick == 2



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import random
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import contextmanager
from operator import attrgetter
from threading import RLock
from typing import Any, Mapping, MutableMapping, MutableSequence, Type

import networkx as nx
from blinker import Signal

from .cache import Cache, TurnEndDict, TurnEndPlanDict, UnitnessCache
from .exc import NotInKeyframeError, TotalKeyError
from .graph import DiGraph, Edge, Node
from .types import CharName, Key, NodeName
from .util import (
	AbstractCharacter,
	AbstractEngine,
	AbstractThing,
	SignalDict,
	getatt,
	timer,
)
from .wrap import MutableMappingUnwrapper
from .xcollections import CompositeDict


class FacadeEntity(MutableMapping, Signal, ABC):
	exists = True
	character = getatt("graph")

	@property
	def rulebook(self):
		if "rulebook" in self._patch:
			return self._patch["rulebook"]
		return self._real.rulebook

	@rulebook.setter
	def rulebook(self, rbname):
		self._patch["rulebook"] = rbname

	@abstractmethod
	def _get_real(self, name):
		raise NotImplementedError()

	def __init__(self, mapping, real_or_name=None, **kwargs):
		self.facade = self.graph = getattr(mapping, "facade", mapping)
		self._mapping = mapping
		is_name = not hasattr(real_or_name, "name") and not hasattr(
			real_or_name, "orig"
		)
		if is_name:
			try:
				self._real = self._get_real(real_or_name)
			except (KeyError, AttributeError):
				pass  # Entity created for Facade. No underlying real entity.
		else:
			self._real = real_or_name
		self._patch = {
			k: v.unwrap() if hasattr(v, "unwrap") else v
			for (k, v) in kwargs.items()
		}

	def __contains__(self, item):
		patch = self._patch
		if item in patch:
			return patch[item] is not ...
		if hasattr(self, "_real"):
			return item in self._real
		return False

	def __iter__(self):
		patch = self._patch
		ks = patch.keys()
		if hasattr(self, "_real"):
			ks |= self._real.keys()
		for k in ks:
			if k not in patch or patch[k] is not ...:
				yield k

	def __len__(self):
		n = 0
		for _ in self:
			n += 1
		return n

	def __getitem__(self, k):
		if k in self._patch:
			if self._patch[k] is ...:
				raise KeyError("{} has been masked.".format(k))
			return self._patch[k]
		if not hasattr(self, "_real"):
			raise KeyError(f"{k} unset, and no underlying Thing")
		ret = self._real[k]
		if hasattr(ret, "unwrap"):  # a wrapped mutable object from the
			# allegedb.wrap module
			ret = ret.unwrap()
			self._patch[k] = ret  # changes will be reflected in the
		# facade but not the original
		return ret

	@abstractmethod
	def _set_plan(self, k, v):
		raise NotImplementedError()

	def __setitem__(self, k, v):
		if k == "name":
			raise KeyError("Can't change names")
		if hasattr(v, "unwrap"):
			v = v.unwrap()
		if self.character.engine._planning:
			return self._set_plan(k, v)
		self._patch[k] = v

	def __delitem__(self, k):
		self._patch[k] = ...

	def apply(self):
		self._real.update(self._patch)
		self._patch = {}

	def unwrap(self):
		return {
			k: v.unwrap() if hasattr(v, "unwrap") else v
			for (k, v) in self.items()
		}


getname = attrgetter("name")


class FacadeEntityMapping(MutableMappingUnwrapper, Signal, ABC):
	"""Mapping that contains entities in a Facade.

	All the entities are of the same type, ``facadecls``, possibly
	being distorted views of entities of the type ``innercls``.

	"""

	facadecls: Type[FacadeEntity]

	@abstractmethod
	def _get_inner_map(self):
		raise NotImplementedError("Missing _get_inner_map")

	def _make(self, k, v):
		if isinstance(v, dict):
			for badkey in ("character", "engine", "name"):
				if badkey in v:
					del v[badkey]
			return self.facadecls(self, k, **v)
		return self.facadecls(self, v)

	engine = getatt("facade.engine")

	def __init__(self, facade, _=None):
		"""Store the facade."""
		super().__init__()
		self.facade = facade
		self._patch = {}

	def __contains__(self, k):
		if k in self._patch:
			return self._patch[k] is not ...
		return k in self._get_inner_map()

	def __iter__(self):
		seen = set()
		for k in self._patch:
			if k not in seen and self._patch[k] is not ...:
				yield k
			seen.add(k)
		for k in self._get_inner_map():
			if k not in seen:
				yield k

	def __len__(self):
		n = 0
		for k in self:
			n += 1
		return n

	def __getitem__(self, k):
		if k not in self and not self.engine._mockup:
			raise KeyError(k)
		if k not in self._patch:
			inner = self._get_inner_map()
			if k in inner:
				v = inner[k]
			else:
				v = {}
			self._patch[k] = self._make(k, v)
		ret = self._patch[k]
		if ret is ...:
			raise KeyError(k)
		if type(ret) is not self.facadecls:
			ret = self._patch[k] = self._make(k, ret)
		return ret

	def __setitem__(self, k, v):
		if not isinstance(v, self.facadecls):
			v = self._make(k, v)
		self._patch[k] = v
		if self is not self.facade.node:
			self.facade.node.send(self, key=k, value=v)

	def __delitem__(self, k):
		if k not in self:
			raise KeyError("{} not present".format(k))
		that = self[k]
		# Units don't work when we're wrapping an EngineProxy or nothing.
		# I'll fix that at some point I guess.
		# 2025-02-06
		if hasattr(self.facade.engine._real, "_unitness_cache") and hasattr(
			that, "users"
		):
			user: CharacterFacade
			for user in list(that.users()):
				user.remove_unit(self.facade.name, k)
		self._patch[k] = ...


class FacadeRulebook(MutableSequence, ABC):
	name: Any
	_fake: list
	engine: "EngineFacade"

	def __iter__(self):
		return iter(self._fake)

	def __getitem__(self, item):
		_ = self._fake[item]
		return FacadeRule(self.engine, item)

	def __setitem__(self, index, value):
		name = getattr(value, "name", value)
		self._fake[index] = name

	def __delitem__(self, index):
		del self._fake[index]

	def __len__(self):
		return len(self._fake)


class FacadeRule:
	class FakeFuncList(MutableSequence):
		def __init__(self, rule, typ):
			self._rule = rule
			self._type = typ

		@property
		def _me(self):
			return getattr(self._rule, f"_fake_{self._type}s")

		def __iter__(self):
			return iter(self._me)

		def __len__(self):
			return len(self._me)

		def __getitem__(self, item):
			realeng = self._rule._engine._real
			return getattr(realeng, self._type)

		def __setitem__(self, key, value):
			self._me[key] = getattr(value, "name", value)

		def __delitem__(self, key):
			del self._me[key]

	def __init__(self, engine, name):
		self._engine = engine
		self.name = name
		realeng = engine._real
		realrule = realeng.rule[name]
		self._fake_triggers = list(map(getname, realrule.triggers))
		self._fake_prereqs = list(map(getname, realrule.prereqs))
		self._fake_actions = list(map(getname, realrule.actions))
		self.triggers = self.FakeFuncList(self, "trigger")
		self.prereqs = self.FakeFuncList(self, "prereq")
		self.actions = self.FakeFuncList(self, "action")

	def apply(self):
		realeng = self._engine._real
		realrule = realeng.rule[self.name]
		realtrigs = list(map(getname, realrule.triggers))
		if self._fake_triggers != realtrigs:
			realrule.triggers = self._fake_triggers
		realpreqs = list(map(getname, realrule.prereqs))
		if self._fake_prereqs != realpreqs:
			realrule.prereqs = self._fake_prereqs
		realacts = list(map(getname, realrule.actions))
		if self._fake_actions != realacts:
			realrule.actions = self._fake_actions


class FacadeNode(FacadeEntity, Node):
	class FacadeNodeUser(Mapping):
		__slots__ = ("_entity",)

		@property
		def only(self):
			if len(self) != 1:
				raise AttributeError("No user, or more than one")
			return self[next(iter(self))]

		def __init__(self, node):
			self._entity = node

		def __iter__(self):
			engine = self._entity.engine
			charn = self._entity.character.name
			return engine._unitness_cache.leader_cache.iter_keys(
				charn, self._entity.name, *engine._btt()
			)

		def __len__(self):
			engine = self._entity.engine
			charn = self._entity.character.name
			return engine._unitness_cache.leader_cache.count_keys(
				charn, self._entity.name, *engine._btt()
			)

		def __contains__(self, item):
			engine = self._entity.engine
			charn = self._entity.character.name
			try:
				return bool(
					engine._unitness_cache.leader_cache.retrieve(
						charn, self._entity.name, item, *engine._btt()
					)
				)
			except KeyError:
				return False

		def __getitem__(self, item):
			if item not in self:
				raise KeyError("Not used by that character", item)
			engine = self._entity.engine
			return engine.character[item]

	class FacadeNodeContent(Mapping):
		__slots__ = ("_entity",)

		def __init__(self, node):
			self._entity = node

		def __iter__(self):
			if hasattr(self._entity, "engine") and hasattr(
				self._entity.engine, "_node_contents_cache"
			):
				# The real contents cache is wrapped by the facade engine.
				try:
					return self._entity.engine._node_contents_cache.retrieve(
						self._entity.character.name,
						self._entity.name,
						*self._entity.engine._btt(),
					)
				except KeyError:
					return
			char = self._entity.character
			myname = self._entity.name
			for name, thing in char.thing.items():
				if thing["location"] == myname:
					yield name

		def __len__(self):
			# slow
			return len(set(self))

		def __contains__(self, item):
			return (
				item in self._entity.character.thing
				and self._entity.character.thing[item]["location"]
				== self._entity.name
			)

		def __getitem__(self, item):
			if item not in self:
				raise KeyError("Not contained here", item, self._entity.name)
			return self._entity.character.thing[item]

	@property
	def portal(self):
		return self.facade.portal[self["name"]]

	def successors(self):
		for dest in self.portal:
			yield self.character.place[dest]

	def contents(self):
		return self.content.values()

	def __init__(self, mapping, real_or_name=None, **kwargs):
		self.name = self.node = getattr(real_or_name, "name", real_or_name)
		super().__init__(mapping, real_or_name, **kwargs)

	def __getitem__(self, item):
		if item == "name":
			return self.name
		return super().__getitem__(item)

	@property
	def content(self):
		return self.FacadeNodeContent(self)

	@property
	def user(self):
		return self.FacadeNodeUser(self)

	def users(self):
		return self.user.values()

	def _set_plan(self, k, v):
		self.character.engine._planned[self.character.engine._curplan][
			self.character.engine.turn
		].append((self.character.name, self.name, k, v))


class FacadeThing(FacadeNode, AbstractThing):
	def __init__(self, mapping, real_or_name, **kwargs):
		from .node import Thing

		location = kwargs.get("location")
		if location is None and not (isinstance(real_or_name, Thing)):
			raise TypeError(
				"FacadeThing needs to wrap a real Thing, or have a location of its own"
			)
		super().__init__(mapping, real_or_name, **kwargs)
		self.character.thing._patch[
			getattr(real_or_name, "name", real_or_name)
		] = self

	def _get_real(self, name):
		return self.character.character.thing[name]

	@property
	def location(self):
		return self.facade.node[self["location"]]

	@location.setter
	def location(self, v):
		if isinstance(v, (FacadePlace, FacadeThing)):
			v = v.name
		if v not in self.facade.node:
			raise KeyError("Location {} not present".format(v))
		self["location"] = v

	def delete(self):
		del self.character.thing[self.name]


class FacadePlace(FacadeNode):
	"""Lightweight analogue of Place for Facade use."""

	def __init__(self, mapping, real_or_name, **kwargs):
		from .node import Place

		super().__init__(mapping, real_or_name, **kwargs)
		if not isinstance(real_or_name, Place):
			if real_or_name in mapping._patch:
				real_or_name = mapping._patch[real_or_name]
			else:
				mapping._patch[real_or_name] = self
				return
		self.character.place._patch[real_or_name.name] = self

	def _get_real(self, name):
		return self.character.character.place[name]

	def add_thing(self, name):
		self.facade.add_thing(name, self.name)

	def new_thing(self, name):
		return self.facade.new_thing(name, self.name)

	def delete(self):
		del self.character.place[self.name]


class FacadePortalMapping(FacadeEntityMapping, ABC):
	cls: Type[FacadeEntityMapping]

	def __getitem__(self, node):
		if node not in self:
			raise KeyError("No such node: {}".format(node))
		if node not in self._patch:
			self._patch[node] = self.cls(self.facade, node)
		ret = self._patch[node]
		if ret is ...:
			raise KeyError("masked")
		if type(ret) is not self.cls:
			nuret = self.cls(self.facade, node)
			if type(ret) is dict:
				nuret._patch = ret
			else:
				nuret.update(ret)
			ret = nuret
		return ret


class FacadePortal(FacadeEntity, Edge):
	"""Lightweight analogue of Portal for Facade use."""

	def __init__(self, mapping, other, **kwargs):
		if hasattr(mapping, "orig"):
			self.orig = mapping.orig
			self.dest = other
		else:
			self.dest = mapping.dest
			self.orig = other
		if hasattr(mapping, "facade"):
			facade = mapping.facade
		else:
			facade = mapping
		try:
			super().__init__(
				facade.character.node[self.orig],
				facade.character.node[self.dest],
				**kwargs,
			)
			self._real = facade.character.portal[self.orig][self.dest]
		except (KeyError, AttributeError):
			if self.orig in facade.node:
				origin = facade.node[self.orig]
			else:
				origin = facade.new_place(self.orig)
			if self.dest in facade.node:
				destination = facade.node[self.dest]
			else:
				destination = facade.new_place(self.dest)
			super().__init__(origin, destination, **kwargs)
			self._real = {}

	def __getitem__(self, item):
		if item == "origin":
			return self.orig
		if item == "destination":
			return self.dest
		return super().__getitem__(item)

	def __setitem__(self, k, v):
		if k in ("origin", "destination"):
			raise TypeError("Portals have fixed origin and destination")
		super().__setitem__(k, v)
		self.character.portal._tampered = True
		self.character.portal[self.orig]._tampered = True

	@property
	def origin(self):
		return self.facade.node[self.orig]

	@property
	def destination(self):
		return self.facade.node[self.dest]

	def _get_real(self, name):
		return self.character.character.portal[self._mapping.orig][name]

	def _set_plan(self, k, v):
		self.character.engine._planned[self.character.engine._curplan][
			self.character.engine.turn
		].append((self.character.name, self.orig, self.dest, k, v))

	def delete(self):
		del self.character.portal[self.orig][self.dest]
		self.character.portal._tampered = True
		self.character.portal[self.orig]._tampered = True


class FacadePortalSuccessors(FacadeEntityMapping):
	facadecls = FacadePortal
	innercls: type

	def __init__(self, facade, origname):
		from .portal import Portal

		self.innercls = Portal
		super().__init__(facade, origname)
		self.orig = origname

	def _make(self, k, v):
		return self.facadecls(self, k, **v)

	def _get_inner_map(self):
		try:
			return self.facade.character.portal[self.orig]
		except AttributeError:
			if not hasattr(self, "_inner_map"):
				self._inner_map = SignalDict()
			return self._inner_map


class FacadePortalPredecessors(FacadeEntityMapping):
	facadecls = FacadePortal
	innercls: type

	def __init__(self, facade, destname):
		from .portal import Portal

		self.innercls = Portal
		super().__init__(facade, destname)
		self.dest = destname

	def _make(self, k, v):
		return self.facadecls(self.facade.portal[k], v)

	def _get_inner_map(self):
		try:
			return self.facade.character.preportal[self.dest]
		except AttributeError:
			return {}


class CharacterFacade(AbstractCharacter):
	def __getstate__(self):
		ports = {}
		for o in self.portal:
			if o not in ports:
				ports[o] = {}
			for d in self.portal[o]:
				ports[o][d] = dict(self.portal[o][d])
		things = {k: dict(v) for (k, v) in self.thing.items()}
		places = {k: dict(v) for (k, v) in self.place.items()}
		stats = {
			k: v.unwrap() if hasattr(v, "unwrap") else v
			for (k, v) in self.graph.items()
		}
		return things, places, ports, stats

	def __setstate__(self, state):
		self.character = None
		self.graph = self.StatMapping(self)
		(
			self.thing._patch,
			self.place._patch,
			self.portal._patch,
			self.graph._patch,
		) = state

	def add_places_from(self, seq, **attrs):
		for place in seq:
			self.add_place(place, **attrs)

	def add_things_from(self, seq, **attrs):
		for thing in seq:
			self.add_thing(thing, **attrs)

	def thing2place(self, name):
		self.place[name] = self.thing.pop(name)

	def place2thing(self, name, location):
		it = self.place.pop(name)
		it["location"] = location
		self.thing[name] = it

	def add_portals_from(self, seq, **attrs):
		for it in seq:
			self.add_portal(*it, **attrs)

	def remove_unit(self, a, b=None):
		if b is None:
			if not isinstance(a, FacadeNode):
				raise TypeError("Need a node or character")
			charn = a.character.name
			noden = a.name
		else:
			charn = a
			if isinstance(b, FacadeNode):
				noden = b.name
			else:
				noden = b
		self.engine._unitness_cache.store(
			self.name, charn, noden, *self.engine._btt(), False
		)

	def add_place(self, name, **kwargs):
		self.place[name] = kwargs

	def add_node(self, name, **kwargs):
		"""Version of add_node that assumes it's a place"""
		self.place[name] = kwargs

	def remove_node(self, node):
		"""Version of remove_node that handles place or thing"""
		if node in self.thing:
			del self.thing[node]
		else:
			del self.place[node]

	def remove_place(self, place):
		del self.place[place]

	def remove_thing(self, thing):
		del self.thing[thing]

	def add_thing(self, name, location, **kwargs):
		kwargs["location"] = location
		self.thing[name] = kwargs

	def add_portal(self, orig, dest, **kwargs):
		self.portal[orig][dest] = kwargs
		self.portal[orig]._tampered = True
		self.portal._tampered = True

	def remove_portal(self, origin, destination):
		del self.portal[origin][destination]
		self.portal._tampered = True
		self.portal[origin]._tampered = True

	def add_edge(self, orig, dest, **kwargs):
		"""Wrapper for add_portal"""
		self.add_portal(orig, dest, **kwargs)

	def add_unit(self, a, b=None):
		if b is None:
			if not isinstance(a, FacadeNode):
				raise TypeError("Need a node or character")
			charn = a.character.name
			noden = a.name
		else:
			charn = a
			if isinstance(b, FacadeNode):
				noden = b.name
			else:
				noden = b
		self.engine._unitness_cache.store(
			self.name, charn, noden, *self.engine._btt(), True
		)

	def __init__(self, character=None, engine=None):
		self.character = character
		if isinstance(engine, EngineFacade):
			self.db = engine
		else:
			self.db = EngineFacade(engine or getattr(character, "db", None))
		self._stat_map = self.StatMapping(self)
		self._rb_patch = {}
		if hasattr(character, "name"):
			self.db.character._patch[character.name] = self
			self._name = character.name
		else:
			self._name = character

	@property
	def graph(self):
		return self._stat_map

	@graph.setter
	def graph(self, v):
		self._stat_map.clear()
		self._stat_map.update(v)

	def portals(self):
		for ds in self.portal.values():
			yield from ds.values()

	class UnitGraphMapping(Mapping):
		class UnitMapping(Mapping):
			def __init__(self, character, graph_name):
				self.character = character
				self.graph_name = graph_name

			def __iter__(self):
				for key in self.character.engine._unitness_cache.iter_keys(
					self.character.name,
					self.graph_name,
					*self.character.engine._btt(),
				):
					if key in self:
						yield key

			def __len__(self):
				return self.character.engine._unitness_cache.count_keys(
					self.character.name,
					self.graph_name,
					*self.character.engine._btt(),
				)

			def __contains__(self, item):
				try:
					return self.character.engine._unitness_cache.retrieve(
						self.character.name,
						self.graph_name,
						item,
						*self.character.engine._btt(),
					)
				except KeyError:
					return False

			def __getitem__(self, item):
				if item not in self:
					if not self.character.engine._mockup:
						raise KeyError(
							"Not a unit of this character in this graph",
							item,
							self.character.name,
							self.graph_name,
						)
					self.character.add_unit(
						self.character.engine.character[self.graph_name].node[
							item
						]
					)
				return self.character.engine.character[self.graph_name].node[
					item
				]

		def __init__(self, character):
			self.character = character

		def __iter__(self):
			engine = self.character.engine
			name = self.character.name
			now = self.character.engine._btt()
			for key in engine._unitness_cache.iter_keys(name, *now):
				if key in self:
					yield key

		def __len__(self):
			return self.character.engine._unitness_cache.count_keys(
				self.character.name, *self.character.engine._btt()
			)

		def __contains__(self, item):
			now = self.character.engine._btt()
			name = self.character.name
			engine = self.character.engine
			try:
				engine._unitness_cache.retrieve(name, item, *now)
				return True
			except KeyError:
				return False

		def __getitem__(self, item):
			if item not in self and not self.character.engine._mockup:
				raise KeyError(
					"Character has no units in graph",
					self.character.name,
					item,
				)
			return self.UnitMapping(self.character, item)

	class ThingMapping(FacadeEntityMapping):
		facadecls = FacadeThing
		innercls: type

		def __init__(self, facade, _=None):
			from .node import Thing

			self.innercls = Thing
			super().__init__(facade, _)

		def _get_inner_map(self):
			try:
				return self.facade.character.thing
			except AttributeError:
				return {}

		def patch(self, d: dict):
			places = d.keys() & self.facade.place.keys()
			if places:
				raise KeyError(
					f"Tried to patch places on thing mapping: {places}"
				)
			self.facade.node.patch(d)

	class PlaceMapping(FacadeEntityMapping):
		facadecls = FacadePlace
		innercls: type

		def __init__(self, facade, _=None):
			from .node import Place

			if not isinstance(facade, CharacterFacade):
				raise TypeError("Need CharacterFacade")

			self.innercls = Place
			super().__init__(facade, _)

		def _get_inner_map(self):
			if isinstance(self.facade.character, nx.Graph) and not isinstance(
				self.facade.character, AbstractCharacter
			):
				return self.facade.character._node
			try:
				return self.facade.character.place
			except AttributeError:
				return {}

		def patch(self, d: dict):
			things = d.keys() & self.facade.thing.keys()
			if things:
				raise KeyError(
					f"Tried to patch things on place mapping: {things}"
				)
			self.facade.node.patch(d)

	def ThingPlaceMapping(self, *args):
		return CompositeDict(self.place, self.thing)

	class PortalSuccessorsMapping(FacadePortalMapping):
		cls = FacadePortalSuccessors

		def __contains__(self, item):
			return item in self.facade.node

		def _get_inner_map(self):
			try:
				return self.facade.character._adj
			except AttributeError:
				return {}

	class PortalPredecessorsMapping(FacadePortalMapping):
		cls = FacadePortalPredecessors

		def __contains__(self, item):
			return item in self.facade._node

		def _get_inner_map(self):
			try:
				return self.facade.character.pred
			except AttributeError:
				return {}

	class StatMapping(MutableMappingUnwrapper, Signal):
		def __init__(self, facade):
			super().__init__()
			self.facade = facade
			self._patch = {}

		def __iter__(self):
			seen = set()
			if hasattr(self.facade.character, "graph"):
				for k in self.facade.character.graph:
					if k not in self._patch:
						yield k
						seen.add(k)
			for k, v in self._patch.items():
				if k not in seen and v is not ...:
					yield k

		def __len__(self):
			n = 0
			for k in self:
				n += 1
			return n

		def __contains__(self, k):
			if k in self._patch:
				return self._patch[k] is not ...
			if (
				hasattr(self.facade.character, "graph")
				and k in self.facade.character.graph
			):
				return True
			return False

		def __getitem__(self, k):
			if k not in self._patch and hasattr(
				self.facade.character, "graph"
			):
				ret = self.facade.character.graph[k]
				if not hasattr(ret, "unwrap"):
					return ret
				self._patch[k] = ret.unwrap()
			if self._patch[k] is ...:
				return KeyError("masked", k)
			return self._patch[k]

		def __setitem__(self, k, v):
			if self.facade.engine._planning:
				self.facade.engine._planned[
					self.facade.character.engine._curplan
				][self.facade.engine.turn].append((self.facade.name, k, v))
				return
			self._patch[k] = v

		def __delitem__(self, k):
			self._patch[k] = ...

	def apply(self):
		"""Do all my changes for real in a batch"""
		realchar = self.character
		realstat = realchar.stat
		realthing = realchar.thing
		realplace = realchar.place
		realport = realchar.portal
		realeng = self.engine._real
		for k, v in self.stat._patch.items():
			if v is ...:
				del realstat[k]
			else:
				realstat[k] = v
		self.stat._patch = {}
		for k, v in self.thing._patch.items():
			if v is ...:
				del realthing[k]
			elif k not in realthing:
				if isinstance(v, FacadeThing):
					v = v._patch
				if "name" in v:
					assert v.pop("name") == k
				realchar.add_thing(k, **v)
			else:
				v.apply()
		self.thing._patch = {}
		for k, v in self.place._patch.items():
			if v is ...:
				del realplace[k]
			elif k not in realplace:
				realchar.add_place(k, **v)
			else:
				v.apply()
		self.place._patch = {}
		if getattr(self.portal, "_tampered", False):
			for orig, dests in self.portal._patch.items():
				if not getattr(dests, "_tampered", False):
					continue
				for dest, v in dests.items():
					if v is ...:
						del realport[orig][dest]
					elif orig not in realport or dest not in realport[orig]:
						realchar.add_portal(orig, dest, **v)
					else:
						v.apply()
				del dests._tampered
			del self.portal._tampered
		self.portal._patch = {}


class EngineFacade(AbstractEngine):
	char_cls = CharacterFacade
	thing_cls = FacadeThing
	place_cls = FacadePlace
	portal_cls = FacadePortal

	class FacadeUniversalMapping(MutableMapping):
		def __init__(self, engine: AbstractEngine):
			assert not isinstance(engine, EngineFacade)
			self.engine = engine
			self._patch = {}
			self._deleted = set()

		def _effective_keys(self):
			return (
				self.engine.universal.keys() | self._patch.keys()
			) - self._deleted

		def __iter__(self):
			yield from self._effective_keys()

		def __len__(self):
			return len(self._effective_keys())

		def __contains__(self, item):
			return item not in self._deleted and (
				item in self._patch or item in self.engine.universal
			)

		def __getitem__(self, item):
			if item in self._patch:
				ret = self._patch[item]
				if ret is ...:
					raise KeyError("Universal key deleted", item)
				return ret
			elif item in self.engine.universal:
				return self.engine.universal[item]
			else:
				raise KeyError("No universal key", item)

		def __setitem__(self, key, value):
			self._patch[key] = value
			if value is not ...:
				self._deleted.discard(key)

		def __delitem__(self, key):
			if key not in self.engine.universal:
				raise KeyError("No key to delete", key)
			self._patch[key] = ...
			self._deleted.add(key)

	class FacadeCharacterMapping(Mapping):
		def __init__(self, engine: "EngineFacade"):
			assert isinstance(engine, EngineFacade)
			self.engine = engine
			self._patch = {}

		def __getitem__(self, key, /):
			realeng = self.engine._real
			if realeng and key not in realeng.character:
				raise KeyError("No character", key)
			if key not in self._patch:
				if realeng:
					fac = CharacterFacade(
						realeng.character[key], engine=realeng
					)
				elif self.engine._mockup:
					fac = CharacterFacade(key, engine=self.engine)
				else:
					raise KeyError("No character", key)
				self._patch[key] = fac
			return self._patch[key]

		def __len__(self):
			return len(self.engine.character)

		def __iter__(self):
			return iter(self.engine.character)

		def apply(self):
			for pat in self._patch.values():
				pat.apply()
			self._patch = {}

	class FacadeCache(Cache):
		def __init__(self, cache, name):
			self._created = cache.db._btt()
			super().__init__(cache.db, name)
			self._real = cache

		def retrieve(self, *args, search=False):
			try:
				return super().retrieve(*args, search=search)
			except (NotInKeyframeError, TotalKeyError):
				return self._real.retrieve(*args, search=search)

		def _get_keycache(
			self, parentity, branch, turn, tick, forward: bool = None
		):
			if forward is None:
				forward = self._real.db._forward
			# Find the last effective keycache before the facade was created.
			# Get the additions and deletions since then.
			# Apply those to the keycache and return it.
			kc = set(
				self._real._get_keycache(
					parentity, *self._created, forward=forward
				)
			)
			added, deleted = self._get_adds_dels(
				parentity, branch, turn, tick, stoptime=self._created
			)
			return frozenset((kc | added) - deleted)

	class FacadeUnitnessCache(FacadeCache, UnitnessCache):
		def __init__(self, cache):
			self._created = cache.db._btt()
			UnitnessCache.__init__(self, cache.db, "unitness_cache")
			self.user_cache = EngineFacade.FacadeCache(
				cache.leader_cache, "user_cache"
			)
			self._real = cache

	def __init__(self, real: AbstractEngine | None, mock=False):
		assert not isinstance(real, EngineFacade)
		self._mockup = mock
		if real is not None:
			for alias in (
				"submit",
				"load_at",
				"function",
				"method",
				"trigger",
				"prereq",
				"action",
				"string",
				"log",
				"debug",
				"info",
				"warning",
				"error",
				"critical",
			):
				try:
					setattr(self, alias, getattr(real, alias))
				except AttributeError:
					print(f"{alias} not implemented on {type(real)}")
		self.closed = False
		self._real = real
		self._planning = False
		self._planned = defaultdict(lambda: defaultdict(list))
		self.character = self.FacadeCharacterMapping(self)
		self.universal = self.FacadeUniversalMapping(real)
		self._rando = random.Random()
		self.world_lock = RLock()
		if real is not None:
			self._rando.setstate(real.universal["rando_state"])
			self.branch, self.turn, self.tick = real._btt()
			self._branches_d = real._branches_d.copy()
			self._turn_end = TurnEndDict(self)
			self._turn_end_plan = TurnEndPlanDict(self)
			if not hasattr(real, "is_proxy"):
				self._turn_end.update(real._turn_end)
				self._turn_end_plan.update(real._turn_end_plan)
				self._nodes_cache = self.FacadeCache(
					real._nodes_cache, "nodes_cache"
				)
				self._things_cache = self.FacadeCache(
					real._things_cache, "things_cache"
				)
				self._unitness_cache = self.FacadeUnitnessCache(
					real._unitness_cache
				)
		else:
			self._branches_d = {}
			self._turn_end_plan = {}

	def _get_node(
		self, char: AbstractCharacter | CharName, node: NodeName
	) -> Node:
		return self.character[char].node[node]

	def _btt(self):
		return self.branch, self.turn, self.tick

	def _set_btt(self, branch: str, turn: int, tick: int) -> None:
		(self.branch, self.turn, self.tick) = (branch, turn, tick)

	def _extend_branch(self, branch: str, turn: int, tick: int) -> None:
		if branch in self._branches_d:
			parent, turn_from, tick_from, turn_to, tick_to = self._branches_d[
				branch
			]
			if (turn, tick) > (turn_to, tick_to):
				self._branches_d[branch] = (
					parent,
					turn_from,
					tick_from,
					turn,
					tick,
				)
		else:
			self._branches_d[branch] = None, turn, tick, turn, tick

	def _start_branch(
		self, parent: str, branch: str, turn: int, tick: int
	) -> None:
		self._branches_d[branch] = (parent, turn, tick, turn, tick)
		self._extend_branch(branch, turn, tick)

	def load_at(self, branch: str, turn: int, tick: int) -> None:
		pass

	def turn_end(self, branch: str = None, turn: int = None) -> int:
		if branch is None:
			branch = self.branch
		if turn is None:
			turn = self.turn
		return self._turn_end[branch, turn]

	def turn_end_plan(self, branch: str = None, turn: int = None) -> int:
		if branch is None:
			branch = self.branch
		if turn is None:
			turn = self.turn
		return self._turn_end_plan[branch, turn]

	def _nbtt(self):
		self.tick += 1
		return self._btt()

	@contextmanager
	def batch(self):
		self.info(
			"Facades already batch all changes, so this batch does nothing"
		)
		yield

	@contextmanager
	def plan(self):
		if getattr(self, "_planning", False):
			raise RuntimeError("Already planning")
		self._planning = True
		start_time = self._btt()
		if hasattr(self, "_curplan"):
			self._curplan += 1
		else:
			# Will break if used in a proxy, which I want to do eventually...
			self._curplan = self._real._last_plan + 1
		yield self._curplan
		self._planning = False
		self._set_btt(*start_time)

	def add_character(
		self,
		name: Key,
		data: nx.Graph | DiGraph = None,
		layout: bool = False,
		node: dict = None,
		edge: dict = None,
		**kwargs,
	):
		self.character._patch[name] = char = CharacterFacade(name, self)
		if data:
			char.become(data)
		if node:
			char.node.update(node)
		if edge:
			char.adj.update(edge)

	def apply(self):
		realeng = self._real
		self.character.apply()
		if not getattr(self, "_planned", None):
			return
		# Do I actually need these sorts? Insertion order's preserved...
		for plan_num in sorted(self._planned):
			with (
				timer(
					f"seconds to apply plan {plan_num}",
					logfun=self._real.debug,
				),
				realeng.plan(),
			):  # resets time at end of block
				for turn in sorted(self._planned[plan_num]):
					# Not setting `realeng.turn` the normal way, because that
					# would save the state of the randomizer, which is not
					# relevant here
					realeng._oturn = turn
					for tup in self._planned[plan_num][turn]:
						if len(tup) == 3:
							char, k, v = tup
							realeng.character[char].stat[k] = v
						elif len(tup) == 4:
							char, node, k, v = tup
							realchar = realeng.character[char]
							if node in realchar.node:
								if k is ...:
									realchar.remove_node(node)
								elif k == "location":
									# assume the location really exists, since
									# it did while planning
									now = realeng._nbtt()
									realeng._things_cache.store(
										char, node, *now, v
									)
									realeng.query.set_thing_loc(
										char, node, *now, v
									)
								else:
									realchar.node[node][k] = v
							elif k == "location":
								realchar.add_thing(node, v)
							else:
								realchar.add_place(node, **{k: v})
						elif len(tup) == 5:
							char, orig, dest, k, v = tup
							realchar = realeng.character[char]
							if (
								orig in realchar.portal
								and dest in realchar.portal[orig]
							):
								if k is ...:
									realchar.remove_portal(orig, dest)
								else:
									realchar.portal[orig][dest][k] = v
							else:
								realchar.add_portal(orig, dest, **{k: v})



# This file is part of allegedb, an object-relational mapper for versioned graphs.
# Copyright (c) Zachary Spector. public@zacharyspector.com
import unittest
from copy import deepcopy
from types import SimpleNamespace

import pytest

from lisien.engine import Engine

testkvs = [
	0,
	1,
	10,
	10**10,
	"spam",
	"eggs",
	"ham",
	"💧",
	"🔑",
	"𐦖",
	("spam", "eggs", "ham"),
]
testvs = [["spam", "eggs", "ham"], {"foo": "bar", 0: 1, "💧": "🔑"}]
testdata = []
for k in testkvs:
	for v in testkvs:
		testdata.append((k, v))
	for v in testvs:
		testdata.append((k, v))
testdata.append(("lol", deepcopy(testdata)))


def graph_objects_create_delete(engine: Engine):
	g = engine.new_character("physical")
	assert not engine._node_exists("physical", 0)
	g.add_node(0)
	assert engine._node_exists("physical", 0)
	assert 0 in g
	g.add_node(1)
	assert 1 in g
	g.add_edge(0, 1)
	assert 1 in g.adj[0]
	assert 1 in list(g.adj[0])
	g.add_edge(2, 3)
	assert 2 in g.node
	assert 3 in g.node
	assert 2 in g.adj
	assert 3 in g.adj[2]
	assert 3 in list(g.adj[2])
	assert 2 in g.pred[3]
	g.add_edge(2, 4)
	assert 2 in g.pred[4]
	assert 2 in list(g.pred[4])
	assert 4 in g.adj[2]
	assert 4 in list(g.adj[2])
	del g.pred[4]
	assert len(g.pred[4]) == 0
	assert 4 not in g.adj[2]
	assert 4 not in list(g.adj[2])
	assert 4 in g.node
	assert 0 not in g.adj[1]
	assert 0 not in list(g.adj[1])
	engine.next_turn()
	assert 0 in g
	assert 1 in g
	engine.branch = "physical_no_edge"
	assert 3 in g.node
	assert 0 in g
	assert engine._node_exists("physical", 0)
	assert 1 in g
	assert 1 in g.adj[0]
	assert 1 in list(g.adj[0])
	assert 0 not in g.adj[1]
	assert 0 not in list(g.adj[1])
	g.remove_edge(0, 1)
	assert 0 in g
	assert 1 in g
	assert 1 not in g.adj[0]
	assert 1 not in list(g.adj[0])
	assert 0 in g.adj
	assert 1 not in g.adj[0]
	engine.branch = "physical_triangle"
	assert 3 in g.node
	assert 2 in g
	g.add_edge(0, 1)
	assert 1 in g.adj[0]
	assert 1 in list(g.adj[0])
	g.add_edge(1, 0)
	assert 0 in g.adj[1]
	assert 0 in list(g.adj[1])
	g.add_edge(1, 2)
	g.add_edge(2, 1)
	g.add_edge(2, 0)
	g.add_edge(0, 2)
	assert 2 in g.adj[0]
	assert 2 in list(g.adj[0])
	engine.branch = "physical_square"
	assert engine._node_exists("physical", 0)
	assert 3 in g.node
	assert 2 in list(g.adj[0])
	assert 2 in g.adj[0]
	engine.next_turn()
	assert 2 in g
	assert 2 in list(g.node.keys())
	assert 2 in list(g.adj[0])
	assert 2 in g.adj[0]
	assert 2 in list(g.adj[0])
	assert engine._node_exists("physical", 0)
	g.remove_edge(2, 0)
	assert 0 not in g.adj[2]
	assert 0 not in list(g.adj[2])
	assert 0 in g.node
	assert engine._node_exists("physical", 0)
	assert 0 not in g.adj[3]
	g.add_edge(3, 0)
	assert 0 in g.adj[3]
	assert engine.turn == 2
	assert 0 in g.adj[3]
	assert 0 in list(g.adj[3])
	assert 0 in g.node
	assert engine._node_exists("physical", 0)
	assert 2 in g.pred[3]
	assert 3 in g.pred[0]
	engine.branch = "physical_de_edge"
	assert 3 in g.node
	assert 0 in g.node
	assert engine._node_exists("physical", 0)
	assert 3 in g.adj
	assert 0 in g.adj[3]
	g.remove_node(3)
	assert 3 not in g.node
	assert 3 not in g.adj
	assert 3 not in g.adj[2]
	assert 3 not in g.pred
	assert 3 not in g.pred[0]
	engine.branch = "physical_square"
	assert engine.turn == 2
	assert 0 not in g.adj[2]
	assert 0 not in list(g.adj[2])
	assert 0 in g.adj[3]
	assert 0 in list(g.adj[3])
	assert 3 in g.node
	engine.branch = "physical_nothing"
	assert 0 not in g.adj[2]
	assert 0 not in list(g.adj[2])
	assert 0 in g.adj[3]
	assert 0 in list(g.adj[3])
	assert 3 in g.node
	g.remove_nodes_from((0, 1, 2, 3))
	for n in (0, 1, 2, 3):
		assert n not in g.node
		assert n not in g.adj


def test_branch_lineage(engine):
	# I want an analogue of this test for when you're looking up keyframes
	# in parent branches
	graph_objects_create_delete(engine)
	assert engine.is_ancestor_of("trunk", "physical_no_edge")
	assert engine.is_ancestor_of("trunk", "physical_triangle")
	assert engine.is_ancestor_of("trunk", "physical_nothing")
	assert engine.is_ancestor_of("physical_no_edge", "physical_triangle")
	assert engine.is_ancestor_of("physical_square", "physical_nothing")
	assert not engine.is_ancestor_of("physical_nothing", "trunk")
	assert not engine.is_ancestor_of("physical_triangle", "physical_no_edge")
	engine.time = (
		"trunk",
		engine.branch_start_turn("trunk"),
		engine.branch_start_tick("trunk"),
	)
	g = engine.character["physical"]
	assert 0 not in g.node
	assert 1 not in g.node
	assert 0 not in g.edge
	engine.tick = engine.turn_end()
	assert 0 in g.node
	assert 1 in g.node
	assert 0 in g.edge
	assert 1 in g.edge[0]
	engine.turn = 0
	with pytest.raises(ValueError):
		engine.branch = "physical_no_edge"
	engine.turn = engine.branch_start_turn("physical_no_edge")
	engine.branch = "physical_no_edge"
	engine.next_turn()
	assert 0 in g
	assert 0 in list(g.node.keys())
	assert 1 not in g.edge[0]
	assert 0 not in g.edge[1]
	with pytest.raises(KeyError):
		g.edge[0][1]
	engine.branch = "physical_triangle"
	assert 2 in g.node
	for orig in (0, 1, 2):
		for dest in (0, 1, 2):
			if orig == dest:
				continue
			assert orig in g.edge
			assert dest in g.edge[orig]
	engine.branch = "physical_square"
	assert 0 not in g.edge[2]
	with pytest.raises(KeyError):
		g.edge[2][0]
	engine.turn = 2
	assert 3 in g.node
	assert 1 in g.edge[0]
	assert 2 in g.edge[1]
	assert 3 in g.edge[2]
	engine.branch = "physical_nothing"
	for node in (0, 1, 2):
		assert node not in g.node
		assert node not in g.edge
	engine.branch = "trunk"
	engine.turn = 0
	assert 0 in g.node
	assert 1 in g.node
	assert 0 in g.edge
	assert 1 in g.edge[0]


def test_store_value(engine):
	g = engine.new_character("testgraph")
	g.add_node(0)
	g.add_node(1)
	g.add_edge(0, 1)
	n = g.node[0]
	e = g.edge[0][1]
	for k, v in testdata:
		g.graph[k] = v
		assert k in g.graph
		assert g.graph[k] == v
		del g.graph[k]
		assert k not in g.graph
		n[k] = v
		assert k in n
		assert n[k] == v
		del n[k]
		assert k not in n
		e[k] = v
		assert k in e
		assert e[k] == v
		del e[k]
		assert k not in e


def test_store_dict(engine):
	g = engine.new_character("testgraph")
	g.add_node(0)
	g.add_node(1)
	g.add_edge(0, 1)
	n = g.node[0]
	e = g.edge[0][1]
	for entity in (g.graph, n, e):
		entity[0] = {
			"spam": "eggs",
			"ham": {"baked beans": "delicious"},
			"qux": ["quux", "quuux"],
			"clothes": {"hats", "shirts", "pants"},
			"dicts": {"foo": {"bar": "bas"}, "qux": {"quux": "quuux"}},
		}
	engine.next_turn()
	for entity in (g.graph, n, e):
		assert entity[0]["spam"] == "eggs"
		entity[0]["spam"] = "ham"
		assert entity[0]["spam"] == "ham"
		assert entity[0]["ham"] == {"baked beans": "delicious"}
		entity[0]["ham"]["baked beans"] = "disgusting"
		assert entity[0]["ham"] == {"baked beans": "disgusting"}
		assert entity[0]["qux"] == ["quux", "quuux"]
		entity[0]["qux"] = ["quuux", "quux"]
		assert entity[0]["qux"] == ["quuux", "quux"]
		assert entity[0]["clothes"] == {"hats", "shirts", "pants"}
		entity[0]["clothes"].remove("hats")
		assert entity[0]["clothes"] == {"shirts", "pants"}
		assert entity[0]["dicts"] == {
			"foo": {"bar": "bas"},
			"qux": {"quux": "quuux"},
		}
		del entity[0]["dicts"]["foo"]
		entity[0]["dicts"]["qux"]["foo"] = {"bar": "bas"}
		assert entity[0]["dicts"] == {
			"qux": {"foo": {"bar": "bas"}, "quux": "quuux"}
		}
	engine.turn = 0
	for entity in g.graph, n, e:
		assert entity[0]["spam"] == "eggs"
		assert entity[0]["ham"] == {"baked beans": "delicious"}
		assert entity[0]["qux"] == ["quux", "quuux"]
		assert entity[0]["clothes"] == {"hats", "shirts", "pants"}
		assert entity[0]["dicts"] == {
			"foo": {"bar": "bas"},
			"qux": {"quux": "quuux"},
		}


def test_store_list(engine):
	g = engine.new_character("testgraph")
	g.add_node(0)
	g.add_node(1)
	g.add_edge(0, 1)
	n = g.node[0]
	e = g.edge[0][1]
	for entity in g.graph, n, e:
		entity[0] = [
			"spam",
			("eggs", "ham"),
			{"baked beans": "delicious"},
			["qux", "quux", "quuux"],
			{"hats", "shirts", "pants"},
		]
	engine.next_turn()
	for entity in g.graph, n, e:
		assert entity[0][0] == "spam"
		entity[0][0] = "eggplant"
		assert entity[0][0] == "eggplant"
		assert entity[0][1] == ("eggs", "ham")
		entity[0][1] = ("ham", "eggs")
		assert entity[0][1] == ("ham", "eggs")
		assert entity[0][2] == {"baked beans": "delicious"}
		entity[0][2]["refried beans"] = "deliciouser"
		assert entity[0][2] == {
			"baked beans": "delicious",
			"refried beans": "deliciouser",
		}
		assert entity[0][3] == ["qux", "quux", "quuux"]
		assert entity[0][3].pop() == "quuux"
		assert entity[0][3] == ["qux", "quux"]
		assert entity[0][4] == {"hats", "shirts", "pants"}
		entity[0][4].discard("shame")
		entity[0][4].remove("pants")
		entity[0][4].add("sun")
		assert entity[0][4] == {"hats", "shirts", "sun"}
	engine.turn -= 1
	for entity in g.graph, n, e:
		assert entity[0][0] == "spam"
		assert entity[0][1] == ("eggs", "ham")
		assert entity[0][2] == {"baked beans": "delicious"}
		assert entity[0][3] == ["qux", "quux", "quuux"]
		assert entity[0][4] == {"hats", "shirts", "pants"}


def test_store_set(engine):
	g = engine.new_character("testgraph")
	g.add_node(0)
	g.add_node(1)
	g.add_edge(0, 1)
	n = g.node[0]
	e = g.edge[0][1]
	for entity in g.graph, n, e:
		entity[0] = set(range(10))
	engine.next_turn()
	for entity in g.graph, n, e:
		assert entity[0] == set(range(10))
		for j in range(0, 12, 2):
			entity[0].discard(j)
	engine.turn = 0
	for entity in g.graph, n, e:
		assert entity[0] == set(range(10))



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The top level of the lisien world model, the Character.

Based on NetworkX DiGraph objects with various additions and
conveniences.

A Character is a graph that follows rules. Its rules may be assigned
to run on only some portion of it. Each Character has a ``stat`` property that
acts very much like a dictionary, in which you can store game-time-sensitive
data for the rules to use.

You can designate some nodes in one Character as units of another,
and then assign a rule to run on all of a Character's units. This is
useful for the common case where someone in your game has a location
in the physical world (here, a Character, called 'physical') but also
has a behavior flowchart, or a skill tree, that isn't part of the
physical world. In that case, the flowchart is the person's Character,
and their node in the physical world is a unit of it.

"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Mapping
from itertools import chain
from types import MethodType
from typing import TYPE_CHECKING, Iterable, Iterator

import networkx as nx
from blinker import Signal

from .cache import FuturistWindowDict, PickyDefaultDict
from .exc import AmbiguousLeaderError, WorldIntegrityError
from .facade import CharacterFacade
from .graph import (
	DiGraphPredecessorsMapping,
	DiGraphSuccessorsMapping,
	GraphNodeMapping,
)
from .node import Node, Place, Thing
from .portal import Portal
from .query import CharacterStatAlias, UnitsAlias
from .rule import RuleBook
from .rule import RuleFollower as BaseRuleFollower
from .rule import RuleMapping
from .types import (
	CharName,
	EdgeValDict,
	Key,
	NodeName,
	RulebookTypeStr,
	Stat,
	StatDict,
)
from .util import AbstractCharacter, getatt, singleton_get, timer
from .wrap import MutableMappingUnwrapper, SpecialMapping

if TYPE_CHECKING:
	from .engine import Engine


def grid_2d_8graph(m, n):
	"""Make a 2d graph that's connected 8 ways, with diagonals"""
	me = nx.Graph()
	nodes = me.nodes
	add_node = me.add_node
	add_edge = me.add_edge
	for i in range(m):
		for j in range(n):
			add_node((i, j))
			if i > 0:
				add_edge((i, j), (i - 1, j))
				if j > 0:
					add_edge((i, j), (i - 1, j - 1))
			if j > 0:
				add_edge((i, j), (i, j - 1))
			if (i - 1, j + 1) in nodes:
				add_edge((i, j), (i - 1, j + 1))
	return me


class CharRuleMapping(RuleMapping):
	"""Get rules by name, or make new ones by decorator

	You can access the rules in this either dictionary-style or as
	attributes. This is for convenience if you want to get at a rule's
	decorators, eg. to add an Action to the rule.

	Using this as a decorator will create a new rule, named for the
	decorated function, and using the decorated function as the
	initial Action.

	Using this like a dictionary will let you create new rules,
	appending them onto the underlying :class:`RuleBook`; replace one
	rule with another, where the new one will have the same index in
	the :class:`RuleBook` as the old one; and activate or deactivate
	rules. The name of a rule may be used in place of the actual rule,
	so long as the rule already exists.

	You can also set a rule active or inactive by setting it to
	``True`` or ``False``, respectively. Inactive rules are still in
	the rulebook, but won't be followed.

	"""

	def __init__(
		self,
		character: Character,
		rulebook: RuleBook,
		booktyp: RulebookTypeStr,
	):
		"""Initialize as usual for the ``rulebook``, mostly.

		My ``character`` property will be the one passed in, and my
		``_table`` will be the ``booktyp`` with ``"_rules"`` appended.

		"""
		super().__init__(rulebook.engine, rulebook)
		self.character = character
		self._table = booktyp + "_rules"


class RuleFollower(BaseRuleFollower):
	"""Mixin class. Has a rulebook, which you can get a RuleMapping into."""

	character: Character
	engine: "Engine"
	_book: RulebookTypeStr

	def _get_rule_mapping(self):
		return CharRuleMapping(self.character, self.rulebook, self._book)

	@abstractmethod
	def _get_rulebook_cache(self):
		pass

	def _get_rulebook_name(self):
		try:
			return self._get_rulebook_cache().retrieve(
				self.character.name, *self.engine._btt()
			)
		except KeyError:
			ret = (self._book + "_rulebook", self.character.name)
			self._set_rulebook_name(ret)
			return ret

	def _set_rulebook_name(self, n):
		branch, turn, tick = self.engine._nbtt()
		self.engine.query.set_rulebook_on_character(
			self._book + "_rulebook",
			self.character.name,
			branch,
			turn,
			tick,
			n,
		)
		self._get_rulebook_cache().store(
			self.character.name, branch, turn, tick, n
		)

	def __contains__(self, k: Key) -> bool:
		return self.engine._active_rules_cache.contains_key(
			self._get_rulebook_name(), *self.engine._btt()
		)


class Character(AbstractCharacter, RuleFollower):
	"""A digraph that follows game rules and has a containment hierarchy

	Nodes in a Character are subcategorized into Things and
	Places. Things have locations, and those locations may be Places
	or other Things. To get at those, use the `thing` and `place`
	mappings -- but in situations where the distinction does not matter,
	you may simply address the Character as a mapping, as in NetworkX.

	Characters may have units in other Characters. These are just
	nodes. You can apply rules to a Character's units, and thus to
	any collection of nodes you want, perhaps in many different
	Characters. The ``unit`` attribute handles this. It is a mapping,
	keyed by the other Character's name, then by the name of the node
	that is this Character's unit. In the common case where a
	Character has exactly one unit, it may be retrieved as
	``unit.only``. When it has more than one unit, but only has
	any units in a single other Character, you can get the mapping
	of units in that Character as ``unit.node``. Add units with the
	``add_unit`` method and remove them with ``del_unit``.

	You can assign rules to Characters with their ``rule`` attribute,
	typically using it as a decorator (see :mod:`lisien.rule`). You can do the
	same to some of Character's
	attributes:

	* ``thing.rule`` to make a rule run on all Things in this Character
	  every turn
	* ``place.rule`` to make a rule run on all Places in this Character
	  every turn
	* ``node.rule`` to make a rule run on all Things and Places in this
	  Character every turn
	* ``unit.rule`` to make a rule run on all the units this
	  Character has every turn, regardless of what Character the
	  unit is in
	* ``adj.rule`` to make a rule run on all the edges this Character
	  has every turn

	"""

	_book = "character"

	def remove_portal(self, origin: NodeName, destination: NodeName):
		__doc__ = self.remove_edge.__doc__
		super().remove_edge(origin, destination)

	@property
	def character(self):
		return self

	def _get_rulebook_cache(self):
		return self.engine._characters_rulebooks_cache

	def __bool__(self):
		try:
			return (
				self.engine._graph_cache.retrieve(
					self.name, *self.engine._btt()
				)
				!= "Deleted"
			)
		except KeyError:
			return False

	def __repr__(self):
		return "{}.character[{}]".format(repr(self.engine), repr(self.name))

	def __init__(
		self, engine: "Engine", name: CharName, *, init_rulebooks: bool = True
	):
		super().__init__(engine, name)
		self._avatars_cache = PickyDefaultDict(FuturistWindowDict)
		if not init_rulebooks:
			return
		cachemap = {
			"character_rulebook": engine._characters_rulebooks_cache,
			"unit_rulebook": engine._units_rulebooks_cache,
			"character_thing_rulebook": engine._characters_things_rulebooks_cache,
			"character_place_rulebook": engine._characters_places_rulebooks_cache,
			"character_portal_rulebook": engine._characters_portals_rulebooks_cache,
		}
		branch, turn, tick = engine._btt()
		for rulebook, cache in cachemap.items():
			rulebook_name = (rulebook, name)
			engine.query.set_rulebook_on_character(
				rulebook, name, branch, turn, tick, rulebook_name
			)
			cache.store(name, branch, turn, tick, rulebook_name)

	class ThingMapping(
		MutableMappingUnwrapper, SpecialMapping, RuleFollower, Signal
	):
		""":class:`Thing` objects that are in a :class:`Character`"""

		_book = "character_thing"

		engine: "Engine" = getatt("character.engine")
		name: CharName = getatt("character.name")

		def _get_rulebook_cache(self):
			return self.engine._characters_things_rulebooks_cache

		def __init__(self, character: Character):
			"""Store the character and initialize cache."""
			super().__init__(character)
			self.character = character

		def __iter__(self):
			cache = self.engine._things_cache
			char = self.name
			branch, turn, tick = self.engine._btt()
			return cache.iter_things(char, branch, turn, tick)

		def __contains__(self, thing):
			branch, turn, tick = self.engine._btt()
			args = self.character.name, thing, branch, turn, tick
			cache = self.engine._things_cache
			return cache.thing_exists(*args)

		def __len__(self):
			return self.engine._things_cache.count_keys(
				self.character.name, *self.engine._btt()
			)

		def __getitem__(self, thing: NodeName):
			if thing not in self:
				raise KeyError("No such thing: {}".format(thing))
			return self._make_thing(thing)

		def _make_thing(self, thing: NodeName, val: Thing | None = None):
			cache = self.engine._node_objs
			if isinstance(val, Thing):
				th = cache[self.name, thing] = val
			elif (self.name, thing) in cache:
				th = cache[(self.name, thing)]
				if type(th) is not Thing:
					th = cache[self.name, thing] = Thing(self.character, thing)
			else:
				th = cache[(self.name, thing)] = Thing(self.character, thing)
			return th

		def __setitem__(self, thing: NodeName, val: Mapping):
			if not isinstance(val, Mapping):
				raise TypeError("Things are made from Mappings")
			if "location" not in val:
				raise ValueError("Thing needs location")
			val = dict(val)
			branch, turn, tick = self.engine._nbtt()
			self.engine._nodes_cache.store(
				self.character.name, thing, branch, turn, tick, True
			)
			self.engine._things_cache.store(
				self.character.name, thing, branch, turn, tick, val["location"]
			)
			self.engine.query.exist_node(
				self.character.name, thing, branch, turn, tick, True
			)
			self.engine.query.set_thing_loc(
				self.character.name,
				thing,
				branch,
				turn,
				tick,
				val.pop("location"),
			)
			th = self._make_thing(thing)
			th.clear()
			th.update({k: v for (k, v) in val.items() if k != "location"})

		def __delitem__(self, thing: NodeName):
			self[thing].delete()

		def __repr__(self):
			return "{}.character[{}].thing".format(
				repr(self.engine), repr(self.name)
			)

	class PlaceMapping(
		MutableMappingUnwrapper, SpecialMapping, RuleFollower, Signal
	):
		""":class:`Place` objects that are in a :class:`Character`"""

		_book = "character_place"

		def _get_rulebook_cache(self):
			return self.engine._characters_places_rulebooks_cache

		def update(self, __m: dict, **kwargs) -> None:
			self.character.node.update(__m, **kwargs)

		def __init__(self, character: Character):
			"""Store the character."""
			super().__init__(character)
			self.character = character
			self.engine = engine = character.engine
			charn = character.name
			nodes_cache = engine._nodes_cache
			things_cache = engine._things_cache
			iter_nodes = nodes_cache.iter_entities
			nodes_contains = nodes_cache.contains_entity
			things_contains = things_cache.contains_entity
			btt = engine._btt
			self._iter_stuff = (
				iter_nodes,
				nodes_contains,
				things_contains,
				charn,
				btt,
			)
			self._contains_stuff = (
				nodes_contains,
				things_contains,
				charn,
				btt,
			)
			self._get_stuff = self._contains_stuff + (
				engine._node_objs,
				character,
			)
			self._set_stuff = (
				engine._node_exists,
				engine._exist_node,
				engine._get_node,
				charn,
				character,
			)

		def __iter__(self):
			iter_nodes, nodes_contains, things_contains, charn, btt = (
				self._iter_stuff
			)
			branch, turn, tick = btt()
			for node in iter_nodes(charn, branch, turn, tick):
				if nodes_contains(
					charn, node, branch, turn, tick
				) and not things_contains(charn, node, branch, turn, tick):
					yield node

		def __len__(self):
			n = 0
			for n, _ in enumerate(self, start=1):
				pass
			return n

		def __contains__(self, place: NodeName) -> bool:
			nodes_contains, things_contains, charn, btt = self._contains_stuff
			branch, turn, tick = btt()
			return nodes_contains(
				charn, place, branch, turn, tick
			) and not things_contains(charn, place, branch, turn, tick)

		def __getitem__(self, place: NodeName) -> Place:
			(nodes_contains, things_contains, charn, btt, cache, character) = (
				self._get_stuff
			)
			branch, turn, tick = btt()
			if not nodes_contains(
				charn, place, branch, turn, tick
			) or things_contains(charn, place, branch, turn, tick):
				raise KeyError("No such place: {}".format(place))
			if (charn, place) not in cache or not isinstance(
				cache[(charn, place)], Place
			):
				ret = cache[(charn, place)] = Place(character, place)
				return ret
			return cache[(charn, place)]

		def __setitem__(self, place: NodeName, v: Place | StatDict):
			(node_exists, exist_node, get_node, charn, character) = (
				self._set_stuff
			)
			exist_node(charn, place, True)
			pl = get_node(character, place)
			if not isinstance(pl, Place):
				raise KeyError(
					"{} is a {}, not a place".format(place, type(pl).__name__)
				)
			pl.update(v)

		def __delitem__(self, place: NodeName):
			self[place].delete()

		def __repr__(self):
			return "{}.character[{}].place".format(
				repr(self.character.engine), repr(self.character.name)
			)

	class ThingPlaceMapping(GraphNodeMapping, SpecialMapping, Signal):
		"""GraphNodeMapping but for Place and Thing"""

		character: Character = getatt("graph")
		engine: "Engine" = getatt("db")
		name: CharName = getatt("character.name")

		def __init__(self, character: Character):
			"""Store the character."""
			super().__init__(character)
			Signal.__init__(self)
			engine = character.engine
			charn = character.name
			self._contains_stuff = contains_stuff = (
				engine._node_exists,
				charn,
			)
			self._getitem_stuff = contains_stuff + (
				engine._get_node,
				character,
			)
			self._delitem_stuff = contains_stuff + (
				engine._is_thing,
				character.thing,
				character.place,
			)
			self._placemap = character.place

		def __contains__(self, k: NodeName):
			node_exists, charn = self._contains_stuff
			return node_exists(charn, k)

		def __getitem__(self, k: NodeName) -> Thing | Place:
			node_exists, charn, get_node, character = self._getitem_stuff
			if not node_exists(charn, k):
				raise KeyError("No such node: " + str(k))
			return get_node(character, k)

		def __setitem__(self, k: NodeName, v: Place | StatDict):
			self._placemap[k] = v

		def __delitem__(self, k: NodeName):
			(node_exists, charn, is_thing, thingmap, placemap) = (
				self._delitem_stuff
			)
			if not node_exists(charn, k):
				raise KeyError("No such node: " + str(k))
			if is_thing(charn, k):
				del thingmap[k]
			else:
				del placemap[k]

	node_map_cls = ThingPlaceMapping

	class PortalSuccessorsMapping(
		DiGraphSuccessorsMapping, SpecialMapping, RuleFollower
	):
		"""Mapping of nodes that have at least one outgoing edge.

		Maps them to another mapping, keyed by the destination nodes,
		which maps to Portal objects.

		"""

		_book = "character_portal"

		character: Character = getatt("graph")
		engine: "Engine" = getatt("graph.engine")

		def __init__(self, graph):
			super().__init__(graph)
			engine = graph.engine
			charn = graph.name
			self._cporh = engine._characters_portals_rulebooks_cache
			self._getitem_stuff = (engine._node_exists, charn, self._cache)
			self._setitem_stuff = (self._cache, self.Successors)

		def _get_rulebook_cache(self):
			return self._cporh

		def __getitem__(self, orig: NodeName) -> Successors:
			node_exists, charn, cache = self._getitem_stuff
			if node_exists(charn, orig):
				if orig not in cache:
					cache[orig] = self.Successors(self, orig)
				return cache[orig]
			raise KeyError("No such node")

		def __delitem__(self, orig):
			super().__delitem__(orig)

		def update(self, other: EdgeValDict, **kwargs):
			"""Recursively update the stats of all portals

			Input should be a dictionary of dictionaries of dictionaries
			--just like networkx ``DiGraph._edge``.

			This will create portals as needed, but will only delete
			them if you set their value to ``None``. Likewise, stats
			not specified in the input will be left untouched, if they
			are already present, but you can set them to ``None`` to
			delete them.

			"""
			kwargs: EdgeValDict
			engine = self.engine
			planning = engine._planning
			forward = engine._forward
			branch, turn, start_tick = engine._btt()
			exist_edge = engine.query.exist_edge
			edge_val_set = engine.query.edge_val_set
			store_edge = engine._edges_cache.store
			store_edge_val = engine._edge_val_cache.store
			iter_edge_keys = engine._edge_val_cache.iter_entity_keys
			charn = self.character.name
			tick = start_tick + 1
			with timer(
				"seconds spent updating PortalSuccessorsMapping", engine.debug
			):
				for orig, dests in chain(other.items(), kwargs.items()):
					for dest, kvs in dests.items():
						if kvs is ...:
							for k in iter_edge_keys(
								charn,
								orig,
								dest,
								branch,
								turn,
								start_tick,
								forward=forward,
							):
								store_edge_val(
									charn,
									orig,
									dest,
									k,
									branch,
									turn,
									tick,
									...,
									planning=planning,
									forward=forward,
									loading=True,
								)
								edge_val_set(
									charn,
									orig,
									dest,
									k,
									branch,
									turn,
									tick,
									...,
								)
								tick += 1
							store_edge(
								charn,
								orig,
								dest,
								branch,
								turn,
								tick,
								False,
								planning=planning,
								forward=forward,
								loading=True,
							)
							exist_edge(
								charn, orig, dest, branch, turn, tick, False
							)
							tick += 1
						else:
							store_edge(
								charn,
								orig,
								dest,
								branch,
								turn,
								tick,
								True,
								planning=planning,
								forward=forward,
								loading=True,
							)
							exist_edge(
								charn, orig, dest, branch, turn, tick, True
							)
							tick += 1
							for k, v in kvs.items():
								store_edge_val(
									charn,
									orig,
									dest,
									k,
									branch,
									turn,
									tick,
									v,
									planning=planning,
									forward=forward,
									loading=True,
								)
								edge_val_set(
									charn,
									orig,
									dest,
									k,
									branch,
									turn,
									tick,
									v,
								)
								tick += 1
			self.engine.tick = tick

		class Successors(DiGraphSuccessorsMapping.Successors):
			"""Mapping for possible destinations from some node."""

			engine: "Engine" = getatt("graph.engine")

			def __init__(
				self,
				container: Character.PortalSuccessorsMapping,
				orig: NodeName,
			):
				super().__init__(container, orig)
				graph = self.graph
				engine = graph.engine
				self._getitem_stuff = (engine._get_edge, graph, orig)
				self._setitem_stuff = (
					engine._edge_exists,
					engine._exist_edge,
					graph.name,
					orig,
					engine._get_edge,
					graph,
					engine.query.edge_val_set,
					engine._edge_val_cache.store,
					engine._nbtt,
				)

			def __getitem__(self, dest: NodeName) -> Portal:
				get_edge, graph, orig = self._getitem_stuff
				if dest in self:
					return get_edge(graph, orig, dest)
				raise KeyError("No such portal", graph, orig, dest)

			def __setitem__(
				self, dest: NodeName, value: Portal | StatDict | ...
			):
				if value is ...:
					del self[dest]
					return
				(
					edge_exists,
					exist_edge,
					charn,
					orig,
					get_edge,
					graph,
					db_edge_val_set,
					edge_val_cache_store,
					nbtt,
				) = self._setitem_stuff
				exist_edge(charn, orig, dest)
				for k, v in value.items():
					branch, turn, tick = nbtt()
					db_edge_val_set(
						charn, orig, dest, k, branch, turn, tick, v
					)
					edge_val_cache_store(
						charn, orig, dest, k, branch, turn, tick, v
					)

			def __delitem__(self, dest: NodeName):
				if dest not in self:
					raise KeyError("No portal to {}".format(dest))
				self[dest].delete()

			def update(
				self, other: dict[NodeName, StatDict] | None = None, **kwargs
			):
				kwargs: dict[NodeName, StatDict]
				if other is ...:
					it = kwargs.items()
				else:
					it = chain(other.items(), kwargs.items())
				for dest, vs in it:
					self.graph.add_edge(self.orig, dest, **vs)

	adj_cls = PortalSuccessorsMapping

	class PortalPredecessorsMapping(
		DiGraphPredecessorsMapping, SpecialMapping, RuleFollower
	):
		"""Mapping of nodes that have at least one incoming edge.

		Maps to another mapping keyed by the origin nodes, which maps to
		Portal objects.

		"""

		_book = "character_portal"

		def __init__(self, graph: Character):
			super().__init__(graph)
			self._cporc = graph.engine._characters_portals_rulebooks_cache

		def _get_rulebook_cache(self):
			return self._cporc

		class Predecessors(DiGraphPredecessorsMapping.Predecessors):
			"""Mapping of possible origins from some destination."""

			def __init__(
				self,
				container: Character.PortalPredecessorsMapping,
				dest: NodeName,
			):
				super().__init__(container, dest)
				graph = self.graph
				self._setitem_stuff = (
					graph,
					graph.name,
					dest,
					self.db._edge_objs,
				)

			def __setitem__(self, orig: NodeName, value: StatDict):
				graph, graph_name, dest, portal_objs = self._setitem_stuff
				key = (graph_name, orig, dest)
				if key not in portal_objs:
					portal_objs[key] = Portal(graph, orig, dest)
				p = portal_objs[key]
				p.engine._exist_edge(graph_name, orig, dest)
				p.clear()
				p.update(value)

	pred_cls = PortalPredecessorsMapping

	class UnitGraphMapping(SpecialMapping, RuleFollower):
		"""A mapping of other characters in which one has a unit."""

		_book = "unit"

		engine = getatt("character.engine")
		name = getatt("character.name")

		def _get_rulebook_cache(self):
			return self._avrc

		def __init__(self, char: Character):
			"""Remember my character."""
			self.character = char
			self._char_av_cache = {}
			engine = char.engine
			self._avrc = engine._units_rulebooks_cache
			self._add_av = char.add_unit
			avcache = engine._unitness_cache
			get_char_graphs = avcache.iter_char_graphs
			charn = char.name
			btt = engine._btt
			self._iter_stuff = (get_char_graphs, charn, btt)
			self._len_stuff = (avcache.count_graphs, charn, btt)
			self._contains_stuff = (
				avcache.dict_cache.contains_key,
				charn,
				btt,
			)
			self._node_stuff = (
				self._get_char_av_cache,
				avcache.get_char_only_graph,
				charn,
				btt,
			)
			self._only_stuff = (
				avcache.get_char_only_unit,
				charn,
				btt,
				engine._get_node,
				engine.character,
			)

		def __call__(self, unit: Place | Thing):
			"""Add the unit

			It must be an instance of Place or Thing.

			"""
			if unit.__class__ not in (Place, Thing):
				raise TypeError("Only Things and Places may be units")
			self._add_av(unit.name, unit.character.name)

		def __iter__(self):
			"""Iterate over graphs with unit nodes in them"""
			get_char_graphs, charn, btt = self._iter_stuff
			for graph in get_char_graphs(charn, *btt()):
				if graph in self:
					yield graph

		def __contains__(self, k: NodeName):
			retrieve, charn, btt = self._contains_stuff
			got = retrieve(charn, k, *btt())
			return got is not ... and not isinstance(got, Exception)

		def __len__(self):
			"""Number of graphs in which I have a unit."""
			n = 0
			for n, _ in enumerate(self, start=1):
				pass
			return n

		def _get_char_av_cache(self, g: CharName):
			if g not in self:
				raise KeyError
			if g not in self._char_av_cache:
				self._char_av_cache[g] = self.CharacterUnitMapping(self, g)
			return self._char_av_cache[g]

		def __getitem__(self, g):
			return self._get_char_av_cache(g)

		@property
		def node(self) -> dict[NodeName, Place | Thing]:
			"""If I have units in only one graph, return a map of them

			Otherwise, raise AttributeError.

			"""
			get_char_av_cache: MethodType
			get_char_av_cache, get_char_only_graph, charn, btt = (
				self._node_stuff
			)
			try:
				return get_char_av_cache(get_char_only_graph(charn, *btt()))
			except KeyError:
				raise AttributeError(
					"I have no unit, or I have units in many graphs"
				)

		@property
		def only(self) -> Place | Thing:
			"""If I have only one unit, this is it

			Otherwise, raise AttributeError.

			"""
			get_char_only_av, charn, btt, get_node, charmap = self._only_stuff
			try:
				charn, noden = get_char_only_av(charn, *btt())
				return get_node(charmap[charn], noden)
			except (KeyError, TypeError):
				raise AttributeError("I have no unit, or more than one unit")

		class CharacterUnitMapping(Mapping):
			"""Mapping of units of one Character in another Character."""

			def __init__(
				self, outer: Character.UnitGraphMapping, graphn: CharName
			):
				"""Store this character and the name of the other one"""
				self.character = outer.character
				self.engine = engine = outer.engine
				self.name = name = outer.name
				self.graph = graphn
				unitcache = engine._unitness_cache
				btt = engine._btt
				self._iter_stuff = iter_stuff = (
					unitcache.get_char_graph_units,
					unitcache.contains_unit,
					name,
					graphn,
					btt,
				)
				self._contains_stuff = (
					unitcache._base_retrieve,
					name,
					graphn,
					btt,
				)
				get_node = engine._get_node
				self._getitem_stuff = iter_stuff + (
					get_node,
					graphn,
					engine.character,
				)
				self._only_stuff = (
					get_node,
					engine.character,
					graphn,
					engine._btt,
				)

			def __iter__(self):
				"""Iterate over names of unit nodes"""
				get_char_graph_avs, validate, name, graphn, btt = (
					self._iter_stuff
				)
				for unit in get_char_graph_avs(name, graphn, *btt()):
					if validate(name, graphn, unit, *btt()):
						yield unit

			def __contains__(self, av: NodeName):
				base_retrieve, name, graphn, btt = self._contains_stuff
				return (
					base_retrieve(
						(name, graphn, av, *btt()),
						store_hint=False,
						retrieve_hint=False,
					)
					is True
				)

			def __len__(self):
				"""Number of units of this character in that graph"""
				get_char_graph_avs, validate, name, graphn, btt = (
					self._iter_stuff
				)
				n = 0
				for that in get_char_graph_avs(name, graphn, *btt()):
					if validate(name, graphn, that, *btt()):
						n += 1
				return n

			def __getitem__(self, av: NodeName) -> Place | Thing:
				(
					get_char_graph_avs,
					unitness_cache_has,
					name,
					graphn,
					btt,
					get_node,
					graphn,
					charmap,
				) = self._getitem_stuff
				if unitness_cache_has(name, graphn, av, *btt()):
					return get_node(charmap[graphn], av)
				raise KeyError("No unit: {}".format(av))

			@property
			def only(self) -> Place | Thing:
				"""If I have only one unit, return it; else error"""

				get_node, charmap, graphn, btt = self._only_stuff
				mykey = singleton_get(self.keys())
				if mykey is None:
					raise AttributeError(
						"No unit, or more than one",
						self.character.name,
						*self.engine._btt(),
					)
				try:
					return get_node(charmap[graphn], mykey)
				except KeyError as ex:
					raise AttributeError(
						"The unit doesn't exist",
						graphn,
						mykey,
						*self.engine._btt(),
					) from ex

			def __repr__(self):
				return "{}.character[{}].unit".format(
					repr(self.engine), repr(self.name)
				)

	def facade(self) -> CharacterFacade:
		"""Return a temporary copy of this Character

		A Facade looks like its :class:`Character`, but doesn't do any of the
		stuff Characters do to save changes to the database, nor enable
		time travel. This makes it much speedier to work with.

		"""
		return CharacterFacade(self)

	def add_place(self, node_for_adding: NodeName, **attr):
		"""Add a new Place"""
		attr: StatDict
		self.place[node_for_adding] = attr

	def add_places_from(self, seq: Iterable[NodeName], **attrs):
		attrs: StatDict
		for place in seq:
			self.add_place(place, **attrs)

	def remove_place(self, place: NodeName) -> None:
		"""Remove an existing Place"""
		if place in self.place:
			self.remove_node(place)
		else:
			raise KeyError("No such place: {}".format(place))

	def remove_thing(self, thing: NodeName) -> None:
		"""Remove an existing Thing"""
		if thing in self.thing:
			self.remove_node(thing)
		else:
			raise KeyError("No such thing: {}".format(thing))

	def add_thing(self, name: NodeName, location: NodeName, **kwargs):
		"""Make a new Thing and set its location"""
		kwargs: StatDict
		if name in self.thing:
			raise WorldIntegrityError(
				"Already have a Thing named {}".format(name)
			)
		starter = StatDict(self.node_dict_factory())
		starter.update(kwargs)
		if isinstance(location, Node):
			location = location.name
		starter["location"] = location
		self.thing[name] = starter
		if name not in self._succ:
			self._succ[name] = self.adjlist_inner_dict_factory()
			self._pred[name] = self.adjlist_inner_dict_factory()

	def add_things_from(
		self,
		seq: Iterable[
			tuple[NodeName, NodeName] | tuple[NodeName, NodeName, StatDict]
		],
		**attrs,
	) -> None:
		"""Make many new Things"""
		attrs: StatDict
		for tup in seq:
			name = tup[0]
			location = tup[1]
			kwargs = tup[2] if len(tup) > 2 else attrs
			self.add_thing(name, location, **kwargs)

	def place2thing(self, name: NodeName, location: NodeName) -> None:
		"""Turn a Place into a Thing with the given location.

		It will keep all its attached Portals.

		"""
		self.engine._set_thing_loc(self.name, name, location)
		if (self.name, name) in self.engine._node_objs:
			obj = self.engine._node_objs[self.name, name]
			thing = Thing(self, name)
			for port in obj.portal.values():
				port.origin = thing
			for port in obj.preportal.values():
				port.destination = thing
			self.engine._node_objs[self.name, name] = thing

	def thing2place(self, name: NodeName) -> None:
		"""Unset a Thing's location, and thus turn it into a Place."""
		self.engine._set_thing_loc(self.name, name, None)
		if (self.name, name) in self.engine._node_objs:
			thing = self.engine._node_objs[self.name, name]
			place = Place(self, name)
			for port in thing.portals():
				port.origin = place
			for port in thing.preportals():
				port.destination = place
			self.engine._node_objs[self.name, name] = place

	def add_portal(
		self, origin: NodeName, destination: NodeName, **kwargs
	) -> None:
		"""Connect the origin to the destination with a :class:`Portal`.

		Keyword arguments are attributes of the :class:`Portal`.

		"""
		kwargs: StatDict
		if isinstance(origin, Node):
			origin = origin.name
		if origin not in self.place and origin not in self.thing:
			self.add_place(origin)
		if isinstance(destination, Node):
			destination = destination.name
		if destination not in self.place:
			self.add_place(destination)
		branch, turn, tick = self.engine._nbtt()
		self.engine._edges_cache.store(
			self.name, origin, destination, branch, turn, tick, True
		)
		self.engine.query.exist_edge(
			self.name, origin, destination, branch, turn, tick, True
		)
		for k, v in kwargs.items():
			branch, turn, tick = self.engine._nbtt()
			self.engine._edge_val_cache.store(
				self.name, origin, destination, k, branch, turn, tick, v
			)
			self.engine.query.edge_val_set(
				self.name, origin, destination, k, branch, turn, tick, v
			)

	def new_portal(
		self, origin: NodeName, destination: NodeName, **kwargs
	) -> Portal:
		"""Create a portal and return it"""
		kwargs: StatDict
		self.add_portal(origin, destination, **kwargs)
		return self.engine._get_edge(self, origin, destination)

	def add_portals_from(self, seq, **kwargs):
		"""Make portals for a sequence of (origin, destination) pairs

		Actually, triples are acceptable too, in which case the third
		item is a dictionary of stats for the new :class:`Portal`.
		"""
		for tup in seq:
			orig = tup[0]
			dest = tup[1]
			kwarrgs = tup[2] if len(tup) > 2 else kwargs
			self.add_portal(orig, dest, **kwarrgs)

	def add_unit(
		self,
		a: Place | Thing | Character | CharName,
		b: Place | Thing | NodeName | None = None,
	) -> None:
		"""Start keeping track of a unit.

		Units are nodes in other characters that are in some sense part of
		this one. A common example in strategy games is when a general leads
		an army: the general is one :class:`Character`, with a graph
		representing the state of its AI; the battle map is another character;
		and the general's units, though not in the general's
		:class:`Character`, are still under their command, and therefore
		follow rules defined on the general's ``unit`` property.

		"""
		if self.engine._planning:
			raise NotImplementedError(
				"Currently can't add units within a plan"
			)
		if b is None:
			if not isinstance(a, (Place, Thing)):
				raise TypeError(
					"when called with one argument, "
					"it must be a place or thing"
				)
			g = a.character.name
			n = a.name
		else:
			if isinstance(a, Character):
				g = a.name
			elif not isinstance(a, Key):
				raise TypeError(
					"when called with two arguments, "
					"the first is a character or its name"
				)
			else:
				g = a
			if isinstance(b, (Place, Thing)):
				n = b.name
			elif not isinstance(b, Key):
				raise TypeError(
					"when called with two arguments, "
					"the second is a thing/place or its name"
				)
			else:
				n = b
		# This will create the node if it doesn't exist. Otherwise
		# it's redundant but harmless.
		self.engine._exist_node(g, n)
		# Declare that the node is my unit
		branch, turn, tick = self.engine._nbtt()
		self.engine._unitness_cache.store(
			self.name, g, n, branch, turn, tick, True
		)
		self.engine.query.unit_set(self.name, g, n, branch, turn, tick, True)

	def remove_unit(
		self,
		a: Place | Thing | Character | CharName,
		b: Place | Thing | NodeName | None = None,
	):
		"""This is no longer my unit, though it still exists"""
		if self.engine._planning:
			raise NotImplementedError(
				"Currently can't remove units within a plan"
			)
		if b is None:
			if not isinstance(a, Node):
				raise TypeError(
					"In single argument form, "
					"del_unit requires a Node object "
					"(Thing or Place)."
				)
			g = a.character.name
			n = a.name
		else:
			g = a.name if isinstance(a, Character) else a
			n = b.name if isinstance(b, Node) else b
		branch, turn, tick = self.engine._nbtt()
		self.engine._unitness_cache.store(
			self.name, g, n, branch, turn, tick, False
		)
		self.engine.query.unit_set(self.name, g, n, branch, turn, tick, False)

	def portals(self) -> Iterator[Portal]:
		"""Iterate over all portals."""
		for o in self.adj.values():
			yield from o.values()

	def historical(self, stat: Stat) -> UnitsAlias | CharacterStatAlias:
		"""Get a historical view on the given stat

		This functions like the value of the stat, but changes
		when you time travel. Comparisons performed on the
		historical view can be passed to ``engine.turns_when``
		to find out when the comparison held true.

		"""
		if stat == "units":
			return UnitsAlias(entity=self, stat="units", engine=self.engine)
		return CharacterStatAlias(entity=self, stat=stat, engine=self.engine)



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Sickle cell anemia vs. malaria, a classic example of population genetics.

This script will initialize world.db and the game code libraries to run the
simulation described. To view it, run elide from the same directory
as you ran this script from.

"""

import networkx as nx

from lisien import Engine


def install(
	engine,
	n_creatures=5,
	n_sickles=3,
	malaria_chance=0.05,
	mate_chance=0.05,
	mapsize=(1, 1),
	startpos=(0, 0),
):
	"""Natural Selection on Sickle Cell Anemia

	If anyone carries a pair of sickle betaglobin genes, they die of
	sickle cell anemia.

	Individuals with 1x betaglobin, 1x sickle betaglobin are immune to
	malaria.

	"""
	initmap = nx.grid_2d_graph(*mapsize)
	phys = engine.new_character("physical", data=initmap)
	species = engine.new_character(
		"species",
		mate_chance=mate_chance,
		malaria_chance=malaria_chance,
		n_creatures=n_creatures,
	)
	for n in range(0, n_creatures):
		name = "critter" + str(n)
		phys.add_thing(
			name=name,
			location=startpos,
			sickle_a=(n < n_sickles),
			sickle_b=False,
			male=engine.coin_flip(),
			last_mate_turn=-1,
		)
		assert name in phys.thing
		assert name not in phys.place
		assert name in phys.node, "couldn't add node {} to phys.node".format(
			name
		)
		assert hasattr(phys.node[name], "location")
		species.add_unit("physical", name)
		assert hasattr(species.unit["physical"][name], "location")

	# putting dieoff earlier in the code than mate means that dieoff will
	# be followed before mate is
	@species.unit.rule(always=True)
	def dieoff(critter):
		ret = (
			"malaria"
			if not (critter["sickle_a"] or critter["sickle_b"])
			else "anemia"
		)
		critter.delete()
		# assert (critter.name not in critter.character.node)
		return ret

	@species.unit.rule
	def mate(critter):
		"""If I share my location with another critter, attempt to mate"""
		engine = critter.engine
		species = critter.leader.only
		suitors = list(
			oc
			for oc in critter.location.contents()
			if oc["male"] != critter["male"]
		)
		assert len(suitors) > 0
		other_critter = critter.engine.choice(suitors)
		sickles = [
			critter["sickle_a"],
			critter["sickle_b"],
			other_critter["sickle_a"],
			other_critter["sickle_b"],
		]
		engine.shuffle(sickles)
		name = "critter" + str(species.stat["n_creatures"])
		species.stat["n_creatures"] += 1
		engine.character["physical"].add_thing(
			name,
			critter["location"],
			sickle_a=sickles.pop(),
			sickle_b=sickles.pop(),
			male=engine.coin_flip(),
			last_mate_turn=engine.turn,
		)
		species.add_unit("physical", name)
		critter["last_mate_turn"] = other_critter["last_mate_turn"] = (
			engine.turn
		)
		return "mated"

	@mate.trigger
	def once_per_turn(critter):
		return critter["last_mate_turn"] < critter.engine.turn

	@mate.prereq
	def mate_present(critter):
		for oc in critter.location.contents():
			if oc["male"] != critter["male"]:
				return True
		return False

	@mate.prereq
	def in_the_mood(critter):
		return (
			critter.engine.random() < critter.leader.only.stat["mate_chance"]
		)

	@dieoff.prereq
	def sickle2(critter):
		return critter["sickle_a"] and critter["sickle_b"]

	@dieoff.prereq
	def malaria(critter):
		return (
			not (critter["sickle_a"] or critter["sickle_b"])
			and critter.engine.random()
			< critter.leader.only.stat["malaria_chance"]
		)

	# it would make more sense to keep using species.unit.rule, this
	# is just a test
	@phys.thing.rule
	def wander(critter):
		dests = list(critter.character.place.keys())
		dests.remove(critter["location"])
		dest = critter["destination"] = critter.engine.choice(dests)
		critter.travel_to(dest)

	@wander.trigger
	def not_travelling(critter):
		return (
			"destination" not in critter
			or critter["destination"] == critter.location
		)

	@wander.prereq
	def big_map(critter):
		return len(critter.character.place) > 1


def sickle_cell_test(
	engine,
	n_creatures=5,
	n_sickles=3,
	malaria_chance=0.05,
	mate_chance=0.05,
	mapsize=(1, 1),
	startpos=(0, 0),
	turns=100,
):
	install(
		engine,
		n_creatures,
		n_sickles,
		malaria_chance,
		mate_chance,
		mapsize,
		startpos,
	)
	species = engine.character["species"]
	print(
		"Starting with {} creatures, of which {} have "
		"at least one sickle betaglobin.".format(
			len(species.unit["physical"]),
			sum(
				1
				for critter in species.unit["physical"].values()
				if critter["sickle_a"] or critter["sickle_b"]
			),
		)
	)

	for i in range(0, turns):
		malaria_dead = 0
		anemia_dead = 0
		born = 0
		while engine.turn < i:
			r = engine.next_turn()
			if not r:
				continue
			r = r[0]
			if isinstance(r, Exception):
				raise r
			if "malaria" in r:
				malaria_dead += 1
			if "anemia" in r:
				anemia_dead += 1
			if "mated" in r:
				born += 1
		print(
			"On turn {}, {} critters were born; "
			"{} died of malaria, and {} of sickle cell anemia, "
			"leaving {} alive.".format(
				i,
				born,
				malaria_dead,
				anemia_dead,
				len(engine.character["species"].unit["physical"]),
			)
		)
	print(
		"Of the remaining {} creatures, {} have a sickle betaglobin.".format(
			len(species.unit["physical"]),
			sum(
				1
				for critter in species.unit["physical"].values()
				if critter["sickle_a"] or critter["sickle_b"]
			),
		)
	)


if __name__ == "__main__":
	import tempfile

	with tempfile.TemporaryDirectory() as d:
		with Engine(d, random_seed=69106, clear=True) as engine:
			sickle_cell_test(engine)



# Parable of the Polygons is public domain.
# This implementation is part of lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Implementation of Parable of the Polygons http://ncase.me/polygons/"""

from lisien.character import grid_2d_8graph


def install(eng):
	@eng.function
	def cmp_neighbor_shapes(poly, cmp, stat):
		"""Compare the proportion of neighboring polys with the same shape as this one

		Count the neighboring polys that are the same shape as this one, and return how that compares with
		some stat on the poly's user.

		"""
		from operator import attrgetter

		home = poly.location
		similar = 0
		n = 0
		for n, neighbor_home in enumerate(home.neighbors(), 1):
			# assume only 1 poly per home for now; this is faithful to the original
			try:
				neighbor = next(iter(neighbor_home.contents()))
			except StopIteration:
				continue
			if neighbor.leader is poly.leader:
				similar += 1
		if n == 0:
			# You'd always want to move if you had *no* neighbors, I guess
			return True
		return cmp(poly.character.stat[stat], similar / n)

	@eng.rule(neighborhood=1)
	def relocate(poly):
		"""Move to a random unoccupied place"""
		if "unoccupied" not in poly.engine.universal:
			poly.engine.universal["unoccupied"] = [
				place
				for place in poly.character.place.values()
				if not place.content
			]
		unoccupied = poly.engine.universal["unoccupied"]
		newloc = unoccupied.pop(poly.engine.randrange(0, len(unoccupied)))
		while not newloc:  # the unoccupied location may have been deleted
			newloc = unoccupied.pop(poly.engine.randrange(0, len(unoccupied)))
		unoccupied.append(poly.location)
		poly.location = newloc

	@relocate.trigger
	def similar_neighbors(poly):
		"""Trigger when my neighborhood fails to be enough like me"""
		from operator import ge

		return poly.engine.function.cmp_neighbor_shapes(
			poly, ge, "min_sameness"
		)

	@relocate.trigger
	def dissimilar_neighbors(poly):
		"""Trigger when my neighborhood gets too much like me"""
		from operator import lt

		return poly.engine.function.cmp_neighbor_shapes(
			poly, lt, "max_sameness"
		)

	eng.rulebook["parable"] = [relocate]

	physical = eng.new_character(
		"physical",
		min_sameness=0.1,
		max_sameness=0.9,
		_config={
			"min_sameness": {"control": "slider", "min": 0.0, "max": 1.0},
			"max_sameness": {"control": "slider", "min": 0.0, "max": 1.0},
		},
		data=grid_2d_8graph(20, 20),
	)
	square = eng.new_character("square")
	triangle = eng.new_character("triangle")
	square.unit.rulebook = triangle.unit.rulebook = "parable"

	empty = list(physical.place.values())
	eng.shuffle(empty)
	# distribute 30 of each shape randomly among the empty places
	for i in range(1, 31):
		square.add_unit(
			empty.pop().new_thing(
				"square%i" % i, _image_paths=["atlas://polygons/meh_square"]
			)
		)
	for i in range(1, 31):
		triangle.add_unit(
			empty.pop().new_thing(
				"triangle%i" % i,
				_image_paths=["atlas://polygons/meh_triangle"],
			)
		)


if __name__ == "__main__":
	import shutil
	from tempfile import TemporaryDirectory

	from lisien import Engine

	with TemporaryDirectory() as td:
		with Engine(
			td,
			random_seed=69105,
			connect_string=f"sqlite:///{td}/world.sqlite3",
		) as eng:
			install(eng)
		archive_filename = shutil.make_archive("polygons", "zip", td)
		print("Exported to " + str(archive_filename))



from lisien import Engine
from lisien.xcollections import FunctionStore, StringStore


def test_single_plan(serial_engine):
	eng = serial_engine
	assert eng.turn == 0
	g = eng.new_character("graph")
	g.add_node(0)
	eng.next_turn()
	assert eng.turn == 1
	g.add_node(1)
	with eng.plan():
		eng.turn = 2
		g.add_node(2)
		g.node[2]["clever"] = False
		eng.turn = 3
		g.node[2]["funny"] = True
		g.add_node(3)
		eng.turn = 4
		g.node[2]["successful"] = True
	eng.turn = 1
	assert 2 not in g.node
	eng.branch = "b"
	assert 2 not in g.node
	assert 1 in g
	eng.next_turn()
	assert eng.turn == 2
	assert 2 in g.node
	assert set(g.node[2].keys()) == {"clever"}
	eng.next_turn()
	assert eng.turn == 3
	assert g.node[2]["funny"]
	eng.tick = eng.turn_end_plan()
	assert 3 in g
	assert set(g.node[2].keys()) == {"funny", "clever"}
	eng.next_turn()
	assert eng.turn == 4
	assert g.node[2].keys() == {"funny", "clever", "successful"}
	eng.turn = 2
	eng.tick = eng.turn_end_plan()
	eng.branch = "d"
	assert g.node[2].keys() == {"clever"}
	g.node[2]["funny"] = False
	assert g.node[2].keys() == {"funny", "clever"}
	eng.turn = 3
	assert not g.node[2]["funny"]
	assert 3 not in g.node
	eng.turn = 4
	assert g.node[2].keys() == {"funny", "clever"}
	eng.turn = 2
	eng.branch = "trunk"
	eng.turn = 0
	assert 1 not in g.node
	eng.branch = "c"
	eng.turn = 2
	assert 1 not in g.node
	assert 2 not in g.node
	eng.turn = 0
	eng.branch = "trunk"
	eng.turn = 2
	eng.tick = eng.turn_end_plan()
	assert 2 in g.node


def test_multi_plan(serial_engine):
	eng = serial_engine
	g1 = eng.new_character(1)
	g2 = eng.new_character(2)
	with eng.plan():
		g1.add_node(1)
		g1.add_node(2)
		eng.turn = 1
		g1.add_edge(1, 2)
	eng.turn = 0
	with eng.plan():
		g2.add_node(1)
		g2.add_node(2)
		eng.turn = 1
		g2.add_edge(1, 2)
	eng.turn = 0
	# contradict the plan
	eng.tick = eng.turn_end_plan()
	del g1.node[2]
	assert 1 in g2.node
	assert 2 in g2.node
	eng.turn = 1
	eng.tick = eng.turn_end_plan()
	assert 2 not in g1.node
	assert 2 not in g1.edge[1]
	assert 2 in g2.edge[1]


def test_plan_vs_plan(serial_engine):
	eng = serial_engine
	g1 = eng.new_character(1)
	with eng.plan():
		g1.add_node(1)
		g1.add_node(2)
		eng.turn = 1
		g1.add_edge(1, 2)
		g1.add_node(3)
		g1.add_edge(3, 1)
	eng.turn = 0
	with eng.plan():
		g1.add_node(0)  # not a contradiction, just two plans
		g1.add_edge(0, 1)
	eng.turn = 1
	eng.tick = eng.turn_end_plan()
	assert 0 in g1.node
	assert 1 in g1.node
	assert 2 in g1.node
	assert 3 in g1.node
	assert 1 in g1.edge[0]
	assert 2 in g1.edge[1]
	eng.turn = 0
	eng.tick = eng.turn_end_plan()
	with eng.plan():
		del g1.node[2]
	eng.turn = 2
	eng.tick = eng.turn_end_plan()
	assert 3 not in g1.node
	assert 3 not in g1.adj
	assert 0 in g1.node
	assert 1 in g1.adj[0]


def test_save_load_plan(tmp_path, non_null_database):
	if non_null_database == "sqlite":
		connect_str = f"sqlite:///{tmp_path}/world.db"
		path = None
	else:
		connect_str = None
		path = tmp_path
	with Engine(
		path,
		function=FunctionStore(None),
		method=FunctionStore(None),
		trigger=FunctionStore(None),
		prereq=FunctionStore(None),
		action=FunctionStore(None),
		string={},
		workers=0,
		connect_string=connect_str,
	) as orm:
		g1 = orm.new_character(1)
		g2 = orm.new_character(2)
		with orm.plan():
			g1.add_node(1)
			g1.add_node(2)
			orm.turn = 1
			g1.add_edge(1, 2)
		orm.turn = 0
		with orm.plan():
			g2.add_node(1)
			g2.add_node(2)
			tick2 = orm.tick
			orm.turn = 1
			g2.add_edge(1, 2)
			tick3 = orm.tick
		orm.turn = 0
	with Engine(
		path,
		workers=0,
		function=FunctionStore(None),
		method=FunctionStore(None),
		trigger=FunctionStore(None),
		prereq=FunctionStore(None),
		action=FunctionStore(None),
		string=StringStore({"language": "eng"}, None),
		connect_string=connect_str,
	) as orm:
		g1 = orm.character[1]
		g2 = orm.character[2]
		assert 2 not in g1.node  # because we're before the plan
		# but if we go to after the plan...
		orm.tick = orm.turn_end_plan()
		assert 1 in g1.node
		assert 2 in g1.node
		# contradict the plan
		del g1.node[2]
		assert 1 in g2.node
		assert 2 in g2.node
		orm.next_turn()
		assert orm.turn == 1
		assert 2 not in g1.node
		assert 2 not in g1.edge[1]
		# but, since the stuff that happened in g2 was in a different plan,
		# it still happens
		orm.next_turn()
		assert 1 in g2.node
		assert 2 in g2.node
		assert 2 in g2.edge[1]
	with Engine(
		path,
		workers=0,
		function=FunctionStore(None),
		method=FunctionStore(None),
		trigger=FunctionStore(None),
		prereq=FunctionStore(None),
		action=FunctionStore(None),
		string=StringStore({"language": "eng"}, None),
		connect_string=connect_str,
	) as orm:
		orm.turn = 0
		g1 = orm.character[1]
		g2 = orm.character[2]
		assert 1 in g2.node
		assert 2 in g2.node
		assert 2 not in g1.edge[1]
		assert 2 not in g2.edge[1]
		orm.turn = 1
		assert 2 not in g1.node
		orm.turn = 2
		assert 2 in g2.edge[1]
